# @file edk2_pr_eval
# Checks the diff between a branch and head and then identifies
# if the requested packages need to be built.
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Invocable that checks the diff between a branch and head.

Identifies any packages that needs to be build.

Contains a PrEvalSettingsManager that must be subclassed in a build settings
file. This provides platform specific information to Edk2PrEval invocable
while allowing the invocable itself to remain platform agnostic.
"""
import os
import logging
from io import StringIO
from pathlib import Path
from edk2toolext import edk2_logging
from edk2toolext.invocables.edk2_multipkg_aware_invocable import Edk2MultiPkgAwareInvocable
from edk2toolext.invocables.edk2_multipkg_aware_invocable import MultiPkgAwareSettingsInterface
from edk2toollib.uefi.edk2 import path_utilities
from edk2toollib.uefi.edk2.parsers.dec_parser import DecParser
from edk2toollib.uefi.edk2.parsers.dsc_parser import DscParser
from edk2toollib.uefi.edk2.parsers.inf_parser import InfParser
from edk2toollib.utility_functions import RunCmd


class PrEvalSettingsManager(MultiPkgAwareSettingsInterface):
    """Platform specific Settings for Edk2PrEval.

    provide information necessary for `stuart_pr_eval.exe` or
    `edk2_pr_eval.py` to successfully execute.

    Example: Example: Overriding PrEvalSettingsManager
        ```python
        from edk2toolext.invocables.edk2_pr_eval import PrEvalSettingsManager
        class PrEvalManager(PrEvalSettingsManager):
            def FilterPackagesToTest(self, changedFilesList: list, potentialPackagesList: list) -> list:
                filtered_packages = []
                for file in changedFilesList:
                    for package in potentialPackagesList:
                        if package.startswith(potentialPackagesList):
                            filtered_packages.append(package)

                return list(set(filtered_packages))

            def GetPlatformDscAndConfig(self) -> tuple:
                return None
        ```
    """

    def FilterPackagesToTest(self, changedFilesList: list, potentialPackagesList: list) -> list:
        """Filter potential packages to test based on changed files.

        TIP: Optional Override in a subclass

        Arguments:
            changedFilesList (list): files changed in this PR
            potentialPackagesList (list): packages from `GetPackagesSupported()` or from command line
                option -p, --pkg, --pkg-dir from `Edk2MultiPkgAwareInvocable`

        Returns:
            (list): filtered packages to test
        Note:
            Default implementation does zero filtering
        """
        # default implementation does zero filtering.
        return potentialPackagesList

    def GetPlatformDscAndConfig(self) -> tuple:
        """Provide a platform dsc and config.

        If a platform desires to provide its DSC then Policy 4 will evaluate if
        any of the changes will be built in the dsc.

        TIP: Optional Override in a subclass

        Returns:
            (tuple): (workspace relative path to dsc file, input dictionary of dsc key value pairs)
        """
        return None


class Edk2PrEval(Edk2MultiPkgAwareInvocable):
    """Invocable to determine what packages should be tested.

    Evaluate the changes and determine what packages of the supplied packages should
    be tested based on impact from the changes
    """

    def AddCommandLineOptions(self, parserObj):
        """Adds command line options to the argparser."""
        parserObj.add_argument("--pr-target", dest='pr_target', type=str, default=None,
                               help="PR Branch Target.  Allows build optimizations for pull request"
                               " validation based on files changed. If a package doesn't need testing then it will"
                               " be skipped. Example --pr-target origin/master", required=True)
        parserObj.add_argument("--output-csv-format-string", dest='output_csv_format_string', type=str, default=None,
                               help="Provide format string that will be output to stdout a full csv of packages"
                               " to be tested.  Valid Tokens: {pkgcsv}"
                               " Example --output-csv-format-string test={pkgcsv}")
        parserObj.add_argument("--output-count-format-string", dest='output_count_format_string', type=str,
                               default=None, help="Provide format string that will be output to stdout the count of"
                               " packages to be tested.  Valid Tokens: {pkgcount}"
                               " Example --output-count-format-string PackageCount={pkgcount}")
        super().AddCommandLineOptions(parserObj)

    def RetrieveCommandLineOptions(self, args):
        """Retrieve command line options from the argparser."""
        self.pr_target = args.pr_target
        self.output_csv_format_string = args.output_csv_format_string
        self.output_count_format_string = args.output_count_format_string
        super().RetrieveCommandLineOptions(args)

    def GetVerifyCheckRequired(self):
        """Will not call self_describing_environment.VerifyEnvironment because it might not be set up yet."""
        return False

    def GetSettingsClass(self):
        """Returns the PrEvalSettingsManager class.

        WARNING: PrEvalSettingsManager must be subclassed in your platform settings file.
        """
        return PrEvalSettingsManager

    def GetLoggingFileName(self, loggerType):
        """Returns the filename (PREVALLOG) of where the logs for the Edk2CiBuild invocable are stored in."""
        return "PREVALLOG"

    def Go(self):
        """Executes the core functionality of the Edk2CiBuild invocable."""
        # create path obj for resolving paths.  Since PR eval is run early to determine if a build is
        # impacted by the changes of a PR we must ignore any packages path that are not valid due to
        # not having their submodule or folder populated.
        # A packages path is ok to drop for this because if it isn't populated it is assumed outside
        # the repository and thus will not trigger the build.
        self.edk2_path_obj = path_utilities.Edk2Path(
            self.GetWorkspaceRoot(), self.GetPackagesPath(), error_on_invalid_pp=False)
        self.logger = logging.getLogger("edk2_pr_eval")
        actualPackagesDict = self.get_packages_to_build(self.requested_package_list)

        #
        # Report the packages that need to be built
        #
        self.logger.log(edk2_logging.SECTION, "Output Results")
        self.logger.critical("Need to Build:")
        if len(actualPackagesDict.keys()) > 0:
            max_pkg_width = max(len(t) for t in actualPackagesDict.keys())
            for key, value in actualPackagesDict.items():
                self.logger.critical(f"{key:{max_pkg_width}}  Reason: {value}")
        else:
            self.logger.critical("None")

        #
        # If requested thru cmd line write to std out in defined format
        # This enabled CI pipelines
        #
        if self.output_csv_format_string is not None:
            pkgcsv = ",".join([x for x in actualPackagesDict.keys()])
            print(self.output_csv_format_string.format(pkgcsv=pkgcsv))

        if self.output_count_format_string is not None:
            pkgcount = len(actualPackagesDict.keys())
            print(self.output_count_format_string.format(pkgcount=pkgcount))

        return 0

    def get_packages_to_build(self, possible_packages: list) -> dict:
        """Returns a dictionary of packages to build.

        Arguments:
            possible_packages: list of possible packages

        Returns:
            (dict): filtered packages to build
        """
        self.parsed_dec_cache = {}
        (rc, files) = self._get_files_that_changed_in_this_pr(self.pr_target)
        if rc != 0:
            return {}

        #
        # Policy 0 - A file outside a package was modified and force build
        #            all packages on a modified file outside a package setting
        #            is enabled.
        #
        for f in files:
            if not self.edk2_path_obj.GetContainingPackage(os.path.abspath(f)):
                return dict.fromkeys(possible_packages,
                                     "Policy 0 - Build all packages if "
                                     "a file is modified outside a "
                                     "package.")

        remaining_packages = possible_packages.copy()  # start with all possible packages and remove each
        # package once it is determined to be build.  This optimization
        # avoids checking packages that already will be built.

        packages_to_build = {}  # packages to build.  Key = pkg name, Value = 1st reason for build

        #
        # Policy 1 - CI Settings file defined
        #
        after = self.PlatformSettings.FilterPackagesToTest(files, remaining_packages)
        for a in after:
            if a not in remaining_packages:
                raise ValueError(f"PlatformSettings.FilterPackagesToTest returned package not allowed {a}")
            packages_to_build[a] = "Policy 1 - PlatformSettings - Filter Packages"
            remaining_packages.remove(a)

        # No more packages to eval - return the results
        if len(remaining_packages) == 0:
            return packages_to_build

        #
        # Policy 2: Build any package that has changed
        #
        for f in files:
            try:
                pkg = self.edk2_path_obj.GetContainingPackage(os.path.abspath(f))

            except Exception as e:
                self.logger.warning(f"Failed to get package for file {f}.  Exception {e}")
                # Ignore a file in which we fail to get the package
                continue

            if (pkg not in packages_to_build.keys() and pkg in remaining_packages):
                packages_to_build[pkg] = "Policy 2 - Build any package that has changed"
                remaining_packages.remove(pkg)

        # No more packages to eval - return the results
        if len(remaining_packages) == 0:
            return packages_to_build

        #
        # Policy 3: If a file change is a public file then build all packages that
        #           are dependent on that package.
        #

        # list of packages with public files that have changed
        public_package_changes = []

        # Get all public files in packages
        for f in files:
            try:
                pkg = self.edk2_path_obj.GetContainingPackage(os.path.abspath(f))

            except Exception as e:
                self.logger.warning(f"Failed to get package for file {f}.  Exception {e}")
                # Ignore a file in which we fail to get the package
                continue

            if self._is_public_file(f):
                public_package_changes.append(pkg)
        # de-duplicate list
        public_package_changes = list(set(public_package_changes))

        # Now check all remaining packages to see if they depend on the set of packages
        # with public file changes.
        # NOTE: future enhancement could be to check actual file dependencies
        for a in public_package_changes:
            for p in remaining_packages[:]:  # slice so we can delete as we go
                if (self._does_pkg_depend_on_package(p, a)):
                    packages_to_build[p] = f"Policy 3 - Package depends on {a}"
                    remaining_packages.remove(p)  # remove from remaining packages

        # No more packages to eval - return the results
        if len(remaining_packages) == 0:
            return packages_to_build

        #
        # Policy 4: If a file changed in a module and that module is used in the provided dsc file
        # then the package of the dSC file must be built
        #
        PlatformDscInfo = self.PlatformSettings.GetPlatformDscAndConfig()
        if PlatformDscInfo is not None and len(remaining_packages) > 0:
            if len(remaining_packages) != 1:
                raise Exception("Policy 4 can only be used by builds for a single package")

            # files are all the files changed edk2 workspace root relative path
            changed_modules = self._get_unique_module_infs_changed(files)
            changed_modules = [Path(m) for m in changed_modules]

            # now check DSC
            dsc = DscParser()
            dsc.SetBaseAbsPath(self.edk2_path_obj.WorkspacePath)
            dsc.SetPackagePaths(self.edk2_path_obj.PackagePathList)
            # given that PR eval runs before dependencies are downloaded we must tolerate errors
            dsc.SetNoFailMode()
            dsc.SetInputVars(PlatformDscInfo[1])
            dsc.ParseFile(PlatformDscInfo[0])
            allinfs = dsc.OtherMods + dsc.ThreeMods + dsc.SixMods + dsc.Libs  # get list of all INF files
            allinfs = [Path(i) for i in allinfs]

            #
            # Note: for now we assume that remaining_packages has only 1 package and that it corresponds
            # to the DSC file provided.
            #
            for p in remaining_packages[:]:  # slice so we can delete as we go
                for cm in changed_modules:
                    if cm in allinfs:  # is the changed module listed in the DSC file?
                        packages_to_build[p] = f"Policy 4 - Package Dsc depends on {str(cm)}"
                        remaining_packages.remove(p)  # remove from remaining packages
                        break

        # All done now return result

        return packages_to_build

    def _get_unique_module_infs_changed(self, files: list):
        """Returns a list of edk2 relative paths to modules infs that have changed files."""
        modules = []

        for f in files:
            if os.path.splitext(f) in [".txt", ".md"]:  # ignore markdown and txt files
                continue
            try:
                infs = self.edk2_path_obj.GetContainingModules(os.path.abspath(f))
            except Exception as e:
                self.logger.warning(f"Failed to get module for file {f}. Exception: {str(e)}")
                # ignore errors.  These will occur if a module or last file in folder is deleted as part of the PR
                continue

            if len(infs) > 0:  # if this file is part of any INFs
                modules.extend(infs)

        modules = [self.edk2_path_obj.GetEdk2RelativePathFromAbsolutePath(x) for x in set(modules)]
        logging.debug("Changed Modules: " + str(modules))
        return modules

    def _does_pkg_depend_on_package(self, package_to_eval: str, support_package: str) -> bool:
        """Return if any module in package_to_eval depends on public files defined in support_package."""
        # get filesystem path of package_to_eval
        abs_pkg_path = self.edk2_path_obj.GetAbsolutePathOnThisSystemFromEdk2RelativePath(package_to_eval)

        # loop thru all inf files in the package
        inf_files = self._walk_dir_for_filetypes([".inf"], abs_pkg_path)

        # compare if support_pkg in packages section
        # For each INF file
        for f in inf_files:
            ip = InfParser()
            ip.SetBaseAbsPath(self.edk2_path_obj.WorkspacePath).SetPackagePaths(
                self.edk2_path_obj.PackagePathList).ParseFile(f)

            for p in ip.PackagesUsed:
                if p.startswith(support_package):
                    self.logger.info(f"Module: {f} depends on package {support_package}")
                    return True
        # if never found return False
        return False

    def _get_files_that_changed_in_this_pr(self, base_branch) -> tuple:
        """Get all the files that changed in this pr.

        Returns:
            (int, list[str]): error code, list of files
        """
        # get file differences between pr and base
        output = StringIO()
        cmd_params = f"diff --name-only HEAD..{base_branch}"
        rc = RunCmd("git", cmd_params, outstream=output)

        if (rc == 0):
            self.logger.debug("git diff command returned successfully!")
        else:
            self.logger.critical("git diff returned error return value: %s" % str(rc))
            return (rc, [])

        if (output.getvalue() is None):
            self.logger.info("No files listed in diff")
            return (0, [])

        files = output.getvalue().split()
        for f in files:
            self.logger.debug(f"File Changed: {f}")
        return (0, files)

    def _parse_dec_for_package(self, path_to_package):
        """Find DEC for package and parse it."""
        # find DEC file
        path = None
        try:
            allEntries = os.listdir(path_to_package)
            for entry in allEntries:
                if entry.lower().endswith(".dec"):
                    path = os.path.join(path_to_package, entry)
        except Exception:
            self.logger.warning("Exception: Unable to find DEC for package:{0}".format(path_to_package))
            return None

        if path is None:
            self.logger.warning("Unable to find DEC for package:{0}".format(path_to_package))
            return None

        wsr_dec_path = self.edk2_path_obj.GetEdk2RelativePathFromAbsolutePath(path)

        if path is None or wsr_dec_path == "" or not os.path.isfile(path):
            self.logger.warning("Unable to convert path for DEC for package: {0}".format(path_to_package))
            return None

        # parse it
        dec = DecParser()
        dec.SetBaseAbsPath(self.edk2_path_obj.WorkspacePath).SetPackagePaths(self.edk2_path_obj.PackagePathList)
        dec.ParseFile(wsr_dec_path)
        return dec

    def _is_public_file(self, filepath):
        """Returns if file is a public files."""
        fp = filepath.replace("\\", "/")  # make consistant for easy compare

        self.logger.debug("Is public: " + fp)

        if filepath.lower().endswith(".dec"):  # if DEC file then it is public
            return True

        try:
            pkg = self.edk2_path_obj.GetContainingPackage(os.path.abspath(filepath))
        except Exception as e:
            self.logger.warning(f"Failed to get package for {filepath}.  Exception: {str(e)}")
            return False

        dec = None
        if (pkg in self.parsed_dec_cache):
            dec = self.parsed_dec_cache[pkg]
        else:
            abs_pkg_path = self.edk2_path_obj.GetAbsolutePathOnThisSystemFromEdk2RelativePath(pkg)
            dec = self._parse_dec_for_package(abs_pkg_path)
            self.parsed_dec_cache[pkg] = dec

        if dec is None:
            return False

        for includepath in dec.IncludePaths:  # if in the include path of a package then it is public
            if (pkg + "/" + includepath + "/") in filepath:
                return True

        return False

    def _walk_dir_for_filetypes(self, extensionlist, directory, ignorelist=None):
        """Walks a directory for all items ending in certain extension."""
        if not isinstance(extensionlist, list):
            raise ValueError("Expected list but got " + str(type(extensionlist)))

        if directory is None:
            raise ValueError("No directory given")

        if not os.path.isabs(directory):
            raise ValueError("Directory not abs path")

        if not os.path.isdir(directory):
            raise ValueError("Invalid find directory to walk")

        if ignorelist is not None:
            ignorelist_lower = list()
            for item in ignorelist:
                ignorelist_lower.append(item.lower())

        extensionlist_lower = list()
        for item in extensionlist:
            extensionlist_lower.append(item.lower())

        returnlist = list()
        for Root, Dirs, Files in os.walk(directory):
            for File in Files:
                for Extension in extensionlist_lower:
                    if File.lower().endswith(Extension):
                        ignoreIt = False
                        if (ignorelist is not None):
                            for c in ignorelist_lower:
                                if (File.lower().startswith(c)):
                                    ignoreIt = True
                                    break
                        if not ignoreIt:
                            logging.debug(os.path.join(Root, File))
                            returnlist.append(os.path.join(Root, File))

        return returnlist


def main():
    """Entry point to invoke Edk2PrEval."""
    Edk2PrEval().Invoke()



## @file test_git_dependency.py
# Unit test suite for the GitDependency class.
#
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import unittest
from edk2toolext.environment import var_dict


class TestVarDict(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def test_var_dict_basic_set_get(self):
        v = var_dict.VarDict()
        v.SetValue("test1", "value1", "test 1 comment")
        ## confirm basic get
        vv = v.GetValue("test1")
        self.assertEqual("value1", vv)

    def test_var_dict_get_key_is_none(self):
        v = var_dict.VarDict()
        self.assertIsNone(v.GetValue(None))

    def test_var_dict_get_key_unknown_return_value(self):
        v = var_dict.VarDict()
        self.assertIsNone(v.GetValue("invalidkey"))
        self.assertEqual("test1", v.GetValue("invalidkey", "test1"))

    def test_var_dict_cant_override(self):
        v = var_dict.VarDict()
        v.SetValue("test1", "value1", "test 1 comment")
        ## confirm override == false
        v.SetValue("test1", "value2", "test for override")
        vv = v.GetValue("test1")
        self.assertEqual("value1", vv)

        v.SetValue("test1", "value1", "set same")  # to get coverage
        vv = v.GetValue("test1")
        self.assertEqual("value1", vv)

    def test_var_dict_can_override(self):
        v = var_dict.VarDict()
        v.SetValue("test1", "value1", "test 1 comment", True)
        ## confirm override == true
        v.SetValue("test1", "value2", "test for override")
        vv = v.GetValue("test1")
        self.assertEqual("value2", vv)

    def test_var_dict_can_change_override_state_with_same_set(self):
        v = var_dict.VarDict()
        v.SetValue("test1", "value1", "test 1 comment", True)
        v.SetValue("test1", "value1", "Change override to false", False)
        v.SetValue("test1", "value2", "this should fail", False)
        self.assertNotEqual(v.GetValue("test1"), "value2")

    def test_var_dict_can_change_override_state_with_allow_override(self):
        v = var_dict.VarDict()
        v.SetValue("test1", "value1", "test 1 comment", False)
        v.AllowOverride("test1")
        v.SetValue("test1", "value2", "this should be allowed", False)
        self.assertEqual(v.GetValue("test1"), "value2")
        ## since override was set to False should not change again
        v.SetValue("test1", "value3", "this should not be allowed", False)
        self.assertEqual(v.GetValue("test1"), "value2")

    def test_var_dict_key_not_case_sensitive(self):
        v = var_dict.VarDict()
        v.SetValue("test1", "value1", "test 1 comment")
        ## confirm case sensitivity
        vv = v.GetValue("TEST1")
        self.assertEqual("value1", vv)

    def test_var_dict_key_not_case_sensitive2(self):
        v = var_dict.VarDict()
        v.SetValue("TEST1", "value1", "test 1 comment")
        ## confirm case sensitivity
        vv = v.GetValue("test1")
        self.assertEqual("value1", vv)

    def test_var_dict_key_not_case_sensitive3(self):
        v = var_dict.VarDict()
        v.SetValue("TeSt1", "value1", "test 1 comment")
        ## confirm case sensitivity
        vv = v.GetValue("tEsT1")
        self.assertEqual("value1", vv)

    def test_var_dict_build_value_when_type_para_used(self):
        v = var_dict.VarDict()
        v.SetValue("bld_debug_test1", "build_d_value1", "build debug test 1 comment")
        v.SetValue("bld_release_test1", "build_r_value1", "build release test 1 comment")
        ## confirm with correct build type debug
        vv = v.GetBuildValue("TEST1", "DEBUG")
        self.assertEqual("build_d_value1", vv)
        ## confirm with correct build type release
        vv = v.GetBuildValue("TEST1", "release")
        self.assertEqual("build_r_value1", vv)

    def test_var_dict_build_value_none_for_key(self):
        v = var_dict.VarDict()
        v.SetValue("bld_debug_test1", "build_d_value1", "build test 1 comment")
        self.assertIsNone(v.GetBuildValue(None, "DEBUG"))

    def test_var_dict_build_value_when_type_para_used_wc(self):
        v = var_dict.VarDict()
        v.SetValue("bld_*_test1", "build_value1", "build test 1 comment")
        ## confirm wildcard support build type fail back to *
        vv = v.GetBuildValue("TEST1", "DEBUG")
        self.assertEqual("build_value1", vv)
        vv = v.GetBuildValue("TEST1", "RELEASE")
        self.assertEqual("build_value1", vv)

        ## confirm match has higher priority
        v.SetValue("bld_debug_test1", "build_d_value1", "build test 1 comment")
        vv = v.GetBuildValue("TEST1", "DEBUG")
        self.assertEqual("build_d_value1", vv)

        v.SetValue("bld_release_test1", "build_r_value1", "build test 1 comment")
        vv = v.GetBuildValue("TEST1", "release")
        self.assertEqual("build_r_value1", vv)

        vv = v.GetBuildValue("TEST1", "NOOPT")
        self.assertEqual("build_value1", vv)

    def test_var_dict_build_value_when_target_set(self):
        v = var_dict.VarDict()
        v.SetValue("bld_*_test1", "build_value1", "build test 1 comment")
        v.SetValue("TARGET", "DEBUG", "Set to Debug")
        ## confirm can get it with target set
        vv = v.GetBuildValue("TEST1")
        self.assertEqual("build_value1", vv)

    def test_var_dict_build_value_when_no_build_type(self):
        v = var_dict.VarDict()
        v.SetValue("bld_*_test1", "build_value1", "build test 1 comment")
        ## confirm can't get it without build type or target set
        vv = v.GetBuildValue("TEST1")
        self.assertEqual(None, vv)

    def test_var_dict_get_all_with_no_entires(self):
        v = var_dict.VarDict()
        v.SetValue("test1", "build_value1", "build test 1 comment")
        v.SetValue("test2", "test", "non build value")
        ## confirm result only has 1 value
        vlist = v.GetAllBuildKeyValues("DEBUG")
        self.assertEqual(len(vlist), 0)

    def test_var_dict_get_all_with_no_target(self):
        v = var_dict.VarDict()
        v.SetValue("test1", "build_value1", "build test 1 comment")
        v.SetValue("test2", "test", "non build value")
        ## confirm result only has 1 value
        vlist = v.GetAllBuildKeyValues()
        self.assertEqual(len(vlist), 0)

    def test_var_dict_get_all_build_key_values_and_not_other_values(self):
        v = var_dict.VarDict()
        v.SetValue("bld_*_test1", "build_value1", "build test 1 comment")
        v.SetValue("test2", "test", "non build value")
        ## confirm result only has 1 value
        vlist = v.GetAllBuildKeyValues("DEBUG")
        self.assertEqual(len(vlist), 1)
        self.assertIn("TEST1", vlist.keys())

        ## confirm override behavior
        v.SetValue("Target", "DEBUG", "Set target to debug")
        v.SetValue("bld_release_test1", "build_value1", "build test 1 comment")
        vlist = v.GetAllBuildKeyValues()
        self.assertEqual(len(vlist), 1)
        ## override using parameter for build type
        vlist = v.GetAllBuildKeyValues("RELEASE")
        self.assertEqual(len(vlist), 1)
        self.assertIn("TEST1", vlist.keys())

    def test_var_dict_get_all_nonbuild_key_values(self):
        v = var_dict.VarDict()
        v.SetValue("bld_*_test1", "build_value1", "build test 1 comment")
        v.SetValue("test2", "test", "non build value")
        ## confirm result only has 1 value
        vlist = v.GetAllNonBuildKeyValues()
        self.assertEqual(len(vlist), 1)
        self.assertIn("TEST2", vlist.keys())

    def test_var_dict_print_all(self):
        v = var_dict.VarDict()
        v.SetValue("bld_*_test1", "build_value1", "build test 1 comment")
        v.SetValue("test2", "value1", "test 1 comment overrideable", True)
        v.PrintAll()


if __name__ == '__main__':
    unittest.main()



# @file test_edk2_ci_build.py
# This contains unit tests for the edk2_ci_build
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import unittest
from edk2toolext.invocables.edk2_ci_build import Edk2CiBuild
import sys
import os
import logging
import shutil
from importlib import reload
from edk2toolext.tests.uefi_tree import uefi_tree
from edk2toolext.environment import shell_environment
from edk2toolext.environment import self_describing_environment
from edk2toolext.environment import version_aggregator


class TestEdk2CiBuild(unittest.TestCase):

    minimalTree = None

    def setUp(self):
        TestEdk2CiBuild.restart_logging()
        tree = uefi_tree()
        self.minimalTree = tree.get_workspace()
        print(self.minimalTree)
        pass

    def tearDown(self):
        shell_environment.GetEnvironment().restore_initial_checkpoint()
        buildFolder = os.path.join(self.minimalTree, "Build")
        shutil.rmtree(buildFolder, ignore_errors=True)
        TestEdk2CiBuild.restart_logging()
        # we need to make sure to tear down the version aggregator and the SDE
        self_describing_environment.DestroyEnvironment()
        version_aggregator.ResetVersionAggregator()
        pass

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    @classmethod
    def restart_logging(cls):
        '''
        We restart logging as logging is closed at the end of edk2 invocables.
        We also initialize it at the start.
        Reloading is the easiest way to get fresh state
        '''
        logging.shutdown()
        reload(logging)

    def test_init(self):
        builder = Edk2CiBuild()
        self.assertIsNotNone(builder)

    def test_ci_build(self):
        builder = Edk2CiBuild()
        settings_file = os.path.join(self.minimalTree, "settings.py")
        sys.argv = ["stuart_ci_build", "-c", settings_file]
        try:
            builder.Invoke()
        except SystemExit as e:
            self.assertEqual(e.code, 0, "We should have a non zero error code")
            pass
        self.assertTrue(os.path.exists(os.path.join(self.minimalTree, "Build")))

    def test_merge_config(self):
        descriptor = {'descriptor_file': 'C:\\MyRepo\\.pytool\\Plugin\\MyPlugin1\\MyPlugin1_plug_in.yaml',
                      'module': 'MyPlugin1',
                      'name': 'My Plugin 1',
                      'scope': 'cibuild'}

        global_config = {
            "MyPlugin1": {
                "MySetting1": 'global value 1',
                "MySetting2": 'global value 2',
            },
            "MyPlugin2": {
                "MySetting2": 'global value 2'
            }
        }
        package_config = {
            "MyPlugin1": {
                "MySetting1": 'package value 1',
                "MySetting3": 'package value 3'
            },
            "MyPlugin3": {
                "MySetting3": 'package value 3'
            }
        }
        merged_config = {
            "MySetting1": 'package value 1',
            "MySetting2": 'global value 2',
            "MySetting3": 'package value 3'
        }

        self.assertDictEqual(Edk2CiBuild.merge_config(global_config, {}, descriptor), global_config["MyPlugin1"])
        self.assertDictEqual(Edk2CiBuild.merge_config({}, package_config, descriptor), package_config["MyPlugin1"])
        self.assertDictEqual(Edk2CiBuild.merge_config(global_config, package_config, descriptor), merged_config)



## @file test_omnicache.py
# This contains unit tests for omnicache
#
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import os
import unittest
import logging
import tempfile
import shutil
import yaml
import sys
from io import StringIO
from edk2toolext import omnicache
from edk2toollib import utility_functions


test_dir = None
current_dir = None


def prep_workspace():
    global test_dir, current_dir
    # if test temp dir doesn't exist
    if test_dir is None or not os.path.isdir(test_dir):
        test_dir = tempfile.mkdtemp()
        logging.debug("temp dir is: %s" % test_dir)
    else:
        for _ in range(3):
            try:
                shutil.rmtree(test_dir)
            except OSError as err:
                logging.warning(f"Failed to fully remove {test_dir}: {err}")
            else:
                break
        test_dir = tempfile.mkdtemp()
    current_dir = os.path.abspath(os.getcwd())


def clean_workspace():
    global test_dir, current_dir
    os.chdir(current_dir)
    if test_dir is None:
        return

    if os.path.isdir(test_dir):
        for _ in range(3):
            try:
                shutil.rmtree(test_dir)
            except OSError as err:
                logging.warning(f"Failed to fully remove {test_dir}: {err}")
            else:
                break
        test_dir = None


class TestOmniCache(unittest.TestCase):
    def setUp(self):
        prep_workspace()

    @classmethod
    def setUpClass(cls):
        logger = logging.getLogger('')
        logger.addHandler(logging.NullHandler())
        unittest.installHandler()

    @classmethod
    def tearDownClass(cls):
        clean_workspace()

    def test_omnicache_new(self):
        testcache = os.path.join(os.path.abspath(os.getcwd()), test_dir, "testcache")

        # create a new cache
        oc = omnicache.Omnicache(testcache, create=True, convert=False)

        # check that the new cache was created as a bare repo
        out = StringIO()
        gitret = utility_functions.RunCmd("git", "rev-parse --is-bare-repository", workingdir=testcache, outstream=out)
        assert (gitret == 0)
        assert (out.getvalue().strip().lower() == "true")

        # check that it has the right metadata
        out = StringIO()
        gitret = utility_functions.RunCmd("git",
                                          "config --local omnicache.metadata.version",
                                          workingdir=testcache,
                                          outstream=out)
        assert (gitret == 0)
        assert (out.getvalue().strip().lower() == omnicache.OMNICACHE_VERSION)

        # check that omnicache thinks it's valid
        (valid, _) = oc._ValidateOmnicache()
        assert (valid)

        # remove the metadata to simulate an older omnicache that is candidate for conversion
        gitret = utility_functions.RunCmd("git",
                                          "config --local --unset omnicache.metadata.version",
                                          workingdir=testcache)
        assert (gitret == 0)

        # check that omnicache thinks it's not valid but convertible.
        (valid, convertible) = oc._ValidateOmnicache()
        assert (not valid)
        assert (convertible)

        # attempt to create a new cache - test repo is valid git repo, but doesn't have the meta.
        with self.assertRaises(RuntimeError):
            oc = omnicache.Omnicache(testcache, create=False, convert=False)

        # set the metadata version to an unexpected value
        gitret = utility_functions.RunCmd(
            "git",
            "config --local omnicache.metadata.version {0}".format(omnicache.OMNICACHE_VERSION + "x"),
            workingdir=testcache)
        assert (gitret == 0)

        # check that omnicache thinks it's not valid but convertible.
        (valid, convertible) = oc._ValidateOmnicache()
        assert (not valid)
        assert (convertible)

        # attempt to create a new cache - test repo is valid git repo, but doesn't have the expected meta.
        with self.assertRaises(RuntimeError):
            oc = omnicache.Omnicache(testcache, create=False, convert=False)

        # now make it a non-bare (but technically valid) git repo
        gitret = utility_functions.RunCmd("git", "config --local core.bare false", workingdir=testcache)
        assert (gitret == 0)

        # check that omnicache thinks it's not valid and not convertible.
        (valid, convertible) = oc._ValidateOmnicache()
        assert (not valid)
        assert (not convertible)

        # attempt to create a new cache - test repo is valid non-bare git repo
        with self.assertRaises(RuntimeError):
            oc = omnicache.Omnicache(testcache, create=False, convert=False)

        # make it it a non-valid git repo
        os.remove(os.path.join(testcache, "HEAD"))

        # check that omnicache thinks it's not valid and not convertible.
        (valid, convertible) = oc._ValidateOmnicache()
        assert (not valid)
        assert (not convertible)

        # attempt to create a new cache - test repo is not valid git repo
        with self.assertRaises(RuntimeError):
            oc = omnicache.Omnicache(testcache, create=False, convert=False)

    def test_omnicache_convert(self):
        testcache = os.path.join(os.path.abspath(os.getcwd()), test_dir, "testcache")

        # create a new cache
        oc = omnicache.Omnicache(testcache, create=True, convert=False)

        (valid, _) = oc._ValidateOmnicache()
        assert (valid)

        # remove the metadata to simulate an older omnicache that is candidate for conversion
        gitret = utility_functions.RunCmd("git",
                                          "config --local --unset omnicache.metadata.version",
                                          workingdir=testcache)
        assert (gitret == 0)

        # add an empty file to simulate the old config yaml
        with open(os.path.join(testcache, omnicache.PRE_0_11_OMNICACHE_FILENAME), "w") as yf:
            yf.write("Not A Real YAML File")

        # confirm that _ValidateOmnicache correctly identifies cache state
        (valid, convertible) = oc._ValidateOmnicache()
        assert (not valid)
        assert (convertible)

        # add a traditionally-named remote to the cache to simulate an older omnicache
        gitret = utility_functions.RunCmd(
            "git",
            "remote add pytools-ext https://github.com/tianocore/edk2-pytool-extensions.git",
            workingdir=testcache)
        assert (gitret == 0)

        # add a second copy of the remote remote with a different name to the cache to simulate a duplicate
        gitret = utility_functions.RunCmd(
            "git",
            "remote add pytools-ext2 https://github.com/tianocore/edk2-pytool-extensions.git",
            workingdir=testcache)
        assert (gitret == 0)

        # fetch the remote and its tags to populate the cache for conversion.
        gitret = utility_functions.RunCmd("git", "fetch pytools-ext --tags", workingdir=testcache)
        assert (gitret == 0)

        # confirm that _ValidateOmnicache still correctly identifies cache state
        (valid, convertible) = oc._ValidateOmnicache()
        assert (not valid)
        assert (convertible)

        # re-create the omnicache object to trigger conversion
        oc = omnicache.Omnicache(testcache, create=False, convert=True)

        # validate converted cache
        (valid, _) = oc._ValidateOmnicache()
        assert (valid)

        # verify that old config file was deleted.
        assert (not os.path.exists(os.path.join(testcache, omnicache.PRE_0_11_OMNICACHE_FILENAME)))

        # verify that the traditionally-named remote is no longer in the cache (it should have been renamed with a UUID)
        remotes = omnicache.Omnicache.GetRemotes(testcache)
        assert ("pytools-ext" not in remotes.keys())

        # verify that the URL still is in the cache
        assert ("https://github.com/tianocore/edk2-pytool-extensions.git" in remotes.values())

        # verify that there is exactly one entry in the cache for this URL (duplicates should be removed)
        assert (sum(url == "https://github.com/tianocore/edk2-pytool-extensions.git" for url in remotes.values()) == 1)

        # verify that the former remote name is now the "omnicache display name"
        for (name, url) in remotes.items():
            if (url == "https://github.com/tianocore/edk2-pytool-extensions.git"):
                out = StringIO()
                gitret = utility_functions.RunCmd("git",
                                                  f"config --local omnicache.{name}.displayname",
                                                  workingdir=testcache,
                                                  outstream=out)
                assert (gitret == 0)
                displayname = out.getvalue().strip()
                # if there are duplicates, which displayname is chosen is arbitrary, so allow either.
                assert (displayname == "pytools-ext" or displayname == "pytools-ext2")

        # verify that there are no global tags in the root.
        out = StringIO()
        gitret = utility_functions.RunCmd("git", "tag -l", workingdir=testcache, outstream=out)
        assert (gitret == 0)
        assert (out.getvalue().strip() == "")

        # add an empty file to simulate the old config yaml, but leave metadata in place.
        # this simulates the case where an old version of omnicache ran and updated the remotes
        # with some non-UUID versions, but the metadata version still exists. In this case, conversion
        # should be run to cleanup the remotes.
        with open(os.path.join(testcache, omnicache.PRE_0_11_OMNICACHE_FILENAME), "w") as yf:
            yf.write("Not A Real YAML File")

        # confirm that _ValidateOmnicache correctly identifies cache state
        (valid, convertible) = oc._ValidateOmnicache()
        assert (not valid)
        assert (convertible)

    def test_omnicache_add_remove(self):
        testcache = os.path.join(os.path.abspath(os.getcwd()), test_dir, "testcache")

        # create a new cache
        oc = omnicache.Omnicache(testcache, create=True, convert=False)

        (valid, _) = oc._ValidateOmnicache()
        assert (valid)

        # add a remote with display name
        ret = oc.AddRemote("https://github.com/tianocore/edk2-pytool-extensions.git", name="pytools-ext")
        assert (ret == 0)

        assert (len(omnicache.Omnicache.GetRemotes(testcache).keys()) == 1)
        assert ("https://github.com/tianocore/edk2-pytool-extensions.git"
               in omnicache.Omnicache.GetRemotes(testcache).values())

        # get the remote UUID
        remoteName = oc._LookupRemoteForUrl("https://github.com/tianocore/edk2-pytool-extensions.git")
        assert (remoteName is not None)

        # confirm that remote data is as expected
        remoteData = oc.GetRemoteData()
        assert (len(remoteData.keys()) == 1)
        assert (remoteData[remoteName]["url"] == "https://github.com/tianocore/edk2-pytool-extensions.git")
        assert (remoteData[remoteName]["displayname"] == "pytools-ext")

        # remove the remote and make sure it is gone
        ret = oc.RemoveRemote("https://github.com/tianocore/edk2-pytool-extensions.git")
        assert (ret == 0)
        assert (len(omnicache.Omnicache.GetRemotes(testcache).keys()) == 0)
        assert ("https://github.com/tianocore/edk2-pytool-extensions.git"
               not in omnicache.Omnicache.GetRemotes(testcache).values())

        # add a remote without display name
        ret = oc.AddRemote("https://github.com/tianocore/edk2-pytool-extensions.git")
        assert (ret == 0)

        assert (len(omnicache.Omnicache.GetRemotes(testcache).keys()) == 1)
        assert ("https://github.com/tianocore/edk2-pytool-extensions.git"
               in omnicache.Omnicache.GetRemotes(testcache).values())

        # get the remote UUID
        remoteName = oc._LookupRemoteForUrl("https://github.com/tianocore/edk2-pytool-extensions.git")
        assert (remoteName is not None)

        # confirm that remote data is as expected
        remoteData = oc.GetRemoteData()
        assert (len(remoteData.keys()) == 1)
        assert (remoteData[remoteName]["url"] == "https://github.com/tianocore/edk2-pytool-extensions.git")
        assert ("displayname" not in remoteData[remoteName])

        # add a remote that already exists (with new display name) and make sure it is treated as an update.
        ret = oc.AddRemote("https://github.com/tianocore/edk2-pytool-extensions.git", name="pytools-ext2")
        assert (ret == 0)

        assert (len(omnicache.Omnicache.GetRemotes(testcache).keys()) == 1)
        assert ("https://github.com/tianocore/edk2-pytool-extensions.git"
               in omnicache.Omnicache.GetRemotes(testcache).values())

        # get the remote UUID
        remoteName = oc._LookupRemoteForUrl("https://github.com/tianocore/edk2-pytool-extensions.git")
        assert (remoteName is not None)

        # confirm that remote data is as expected
        remoteData = oc.GetRemoteData()
        assert (len(remoteData.keys()) == 1)
        assert (remoteData[remoteName]["url"] == "https://github.com/tianocore/edk2-pytool-extensions.git")
        assert (remoteData[remoteName]["displayname"] == "pytools-ext2")

        # attempt to remove a non-existent remote
        ret = oc.RemoveRemote("http://thisisnot.com/good.git")
        assert (ret != 0)

    def test_omnicache_update(self):
        testcache = os.path.join(os.path.abspath(os.getcwd()), test_dir, "testcache")

        # create a new cache
        oc = omnicache.Omnicache(testcache, create=True, convert=False)

        (valid, _) = oc._ValidateOmnicache()
        assert (valid)

        # add a remote with display name
        ret = oc.AddRemote("https://github.com/tianocore/edk2-pytool-extensions.git", name="pytools-ext")
        assert (ret == 0)

        assert (len(omnicache.Omnicache.GetRemotes(testcache).keys()) == 1)
        assert ("https://github.com/tianocore/edk2-pytool-extensions.git"
               in omnicache.Omnicache.GetRemotes(testcache).values())

        # get the remote UUID
        remoteName = oc._LookupRemoteForUrl("https://github.com/tianocore/edk2-pytool-extensions.git")
        assert (remoteName is not None)

        # update the URL and displayname of the remote
        ret = oc.UpdateRemote(
            "https://github.com/tianocore/edk2-pytool-extensions.git",
            newUrl="https://github.com/tianocore/edk2-pytool-extensions2.git",
            newName="pytools-ext2")
        assert (ret == 0)

        assert (len(omnicache.Omnicache.GetRemotes(testcache).keys()) == 1)
        assert ("https://github.com/tianocore/edk2-pytool-extensions2.git"
               in omnicache.Omnicache.GetRemotes(testcache).values())

        # make sure UUID didn't change
        assert (remoteName == oc._LookupRemoteForUrl("https://github.com/tianocore/edk2-pytool-extensions2.git"))

        # confirm that remote data is as expected
        remoteData = oc.GetRemoteData()
        assert (len(remoteData.keys()) == 1)
        assert (remoteData[remoteName]["url"] == "https://github.com/tianocore/edk2-pytool-extensions2.git")
        assert (remoteData[remoteName]["displayname"] == "pytools-ext2")

        # update a non-existent URL in the cache and confirm error is returned
        ret = oc.UpdateRemote("https://not.a.real.url.com/git")
        assert (ret != 0)

    def test_omnicache_fetch(self):
        testcache = os.path.join(os.path.abspath(os.getcwd()), test_dir, "testcache")

        # create a new cache
        oc = omnicache.Omnicache(testcache, create=True, convert=False)

        (valid, _) = oc._ValidateOmnicache()
        assert (valid)

        # add a remote with display name
        ret = oc.AddRemote("https://github.com/tianocore/edk2-pytool-extensions.git", name="pytools-ext")
        assert (ret == 0)

        # fetch the remote
        ret = oc.Fetch()
        assert (ret == 0)

        # fetch the remote with 4 jobs
        ret = oc.Fetch(jobs=4)
        assert (ret == 0)

        # get the remote UUID
        remoteName = oc._LookupRemoteForUrl("https://github.com/tianocore/edk2-pytool-extensions.git")
        assert (remoteName is not None)

        # verify that branches were fetched into the omnicache
        assert (len(os.listdir(os.path.join(testcache, "refs", "remotes", remoteName))) != 0)

        # verify that tags were fetched into the omnicache
        assert (len(os.listdir(os.path.join(testcache, "refs", "rtags", remoteName))) != 0)

    def test_omnicache_list(self):
        testcache = os.path.join(os.path.abspath(os.getcwd()), test_dir, "testcache")

        # create a new cache
        oc = omnicache.Omnicache(testcache, create=True, convert=False)

        (valid, _) = oc._ValidateOmnicache()
        assert (valid)

        oc.List()

        # add a remote with display name
        ret = oc.AddRemote("https://github.com/tianocore/edk2-pytool-extensions.git", name="pytools-ext")
        assert (ret == 0)

        oc.List()

    def test_config_files(self):
        testcache = os.path.join(os.path.abspath(os.getcwd()), test_dir, "testcache")
        testyaml = os.path.join(os.path.abspath(os.getcwd()), test_dir, "cfg.yaml")

        # create a new cache
        oc = omnicache.Omnicache(testcache, create=True, convert=False)

        (valid, _) = oc._ValidateOmnicache()
        assert (valid)

        # add a remote with display name
        ret = oc.AddRemote("https://github.com/tianocore/edk2-pytool-extensions.git", name="pytools-ext")
        assert (ret == 0)

        # add a remote with no display name
        ret = oc.AddRemote("https://github.com/tianocore/edk2-pytool-extensions2.git")
        assert (ret == 0)

        # export yaml cfg
        ret = omnicache.Export(oc, testyaml)
        assert (ret == 0)

        # inspect the yaml for correctness
        with open(testyaml) as yf:
            content = yaml.safe_load(yf)

        assert ("remotes" in content)
        assert (len(content["remotes"]) == 2)
        for remote in content["remotes"]:
            if (remote["url"] == "https://github.com/tianocore/edk2-pytool-extensions.git"):
                assert (remote["name"] == "pytools-ext")
            elif (remote["url"] == "https://github.com/tianocore/edk2-pytool-extensions2.git"):
                assert (omnicache.Omnicache._IsValidUuid(remote["name"]))
                # remove the "display name" for input test below
                del remote["name"]
            else:
                # not one of the URLs we populated above = bad.
                assert (remote["url"] not in remote.values())

        # save the yaml file (since we removed one of the displaynames)
        with open(testyaml, "w") as yf:
            yaml.dump(content, yf)

        # remove the remotes
        ret = oc.RemoveRemote("https://github.com/tianocore/edk2-pytool-extensions.git")
        assert (ret == 0)
        ret = oc.RemoveRemote("https://github.com/tianocore/edk2-pytool-extensions2.git")
        assert (ret == 0)

        # confirm we have no remotes
        assert (len(oc.GetRemoteData()) == 0)

        # import yaml cfg
        ret = omnicache.ProcessInputConfig(oc, testyaml)
        assert (ret == 0)

        # check resulting omnicache config
        for remote in oc.GetRemoteData().values():
            if (remote["url"] == "https://github.com/tianocore/edk2-pytool-extensions.git"):
                assert (remote["displayname"] == "pytools-ext")
            elif (remote["url"] == "https://github.com/tianocore/edk2-pytool-extensions2.git"):
                assert ("displayname" not in remote)
            else:
                # not one of the URLs we populated above = bad.
                assert (remote["url"] not in remote.values())

    def test_omnicache_main(self):
        testcache = os.path.join(os.path.abspath(os.getcwd()), test_dir, "testcache")
        # shameless code coverage play
        oldargs = sys.argv
        sys.argv = ["omnicache", "--init", testcache]
        ret = omnicache.main()
        assert (ret == 0)
        sys.argv = ["omnicache", "--new", testcache]
        ret = omnicache.main()
        assert (ret != 0)
        sys.argv = ["omnicache", "--scan", testcache, testcache]
        ret = omnicache.main()
        assert (ret == 0)
        sys.argv = oldargs


if __name__ == '__main__':
    unittest.main()



# @file test_repo_resolver.py
# This contains unit tests for repo resolver
#
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import logging
import os
import unittest
from edk2toolext.environment import repo_resolver
import tempfile


branch_dependency = {
    "Url": "https://github.com/microsoft/mu",
    "Path": "test_repo",
    "Branch": "master"
}

sub_branch_dependency = {
    "Url": "https://github.com/microsoft/mu",
    "Path": "test_repo",
    "Branch": "gh-pages"
}

commit_dependency = {
    "Url": "https://github.com/microsoft/mu",
    "Path": "test_repo",
    "Commit": "b1e35a5d2bf05fb7f58f5b641a702c70d6b32a98"
}
commit_later_dependency = {
    "Url": "https://github.com/microsoft/mu",
    "Path": "test_repo",
    "Commit": "e28910950c52256eb620e35d111945cdf5d002d1"
}

microsoft_commit_dependency = {
    "Url": "https://github.com/Microsoft/microsoft.github.io",
    "Path": "test_repo",
    "Commit": "e9153e69c82068b45609359f86554a93569d76f1"
}
microsoft_branch_dependency = {
    "Url": "https://github.com/Microsoft/microsoft.github.io",
    "Path": "test_repo",
    "Commit": "e9153e69c82068b45609359f86554a93569d76f1"
}

test_dir = None


def prep_workspace():
    global test_dir
    # if test temp dir doesn't exist
    if test_dir is None or not os.path.isdir(test_dir):
        test_dir = tempfile.mkdtemp()
        logging.debug("temp dir is: %s" % test_dir)
    else:
        repo_resolver.clear_folder(test_dir)
        test_dir = tempfile.mkdtemp()


def clean_workspace():
    global test_dir
    if test_dir is None:
        return

    if os.path.isdir(test_dir):
        repo_resolver.clear_folder(test_dir)
        test_dir = None


def get_first_file(folder):
    folder_list = os.listdir(folder)
    for file_path in folder_list:
        path = os.path.join(folder, file_path)
        if os.path.isfile(path):
            return path
    return None


class test_repo_resolver(unittest.TestCase):
    def setUp(self):
        prep_workspace()

    @classmethod
    def setUpClass(cls):
        logger = logging.getLogger('')
        logger.setLevel(logging.DEBUG)
        logger.addHandler(logging.NullHandler())
        unittest.installHandler()

    @classmethod
    def tearDown(cls):
        clean_workspace()

    # check to make sure that we can clone a branch correctly
    def test_clone_branch_repo(self):
        # create an empty directory- and set that as the workspace
        repo_resolver.resolve(test_dir, branch_dependency)
        folder_path = os.path.join(test_dir, branch_dependency["Path"])
        details = repo_resolver.get_details(folder_path)
        self.assertEqual(details['Url'], branch_dependency['Url'])
        self.assertEqual(details['Branch'], branch_dependency['Branch'])

    # don't create a git repo, create the folder, add a file, try to clone in the folder, it should throw an exception
    def test_wont_delete_files(self):
        folder_path = os.path.join(test_dir, commit_dependency["Path"])
        os.makedirs(folder_path)
        file_path = os.path.join(folder_path, "test.txt")
        file_path = os.path.join(
            test_dir, branch_dependency["Path"], "test.txt")
        out_file = open(file_path, "w+")
        out_file.write("Make sure we don't delete this")
        out_file.close()
        self.assertTrue(os.path.isfile(file_path))
        with self.assertRaises(Exception):
            repo_resolver.resolve(test_dir, branch_dependency)
            self.fail("We shouldn't make it here")
        self.assertTrue(os.path.isfile(file_path))

    # don't create a git repo, create the folder, add a file, try to clone in the folder, will force it to happen
    def test_will_delete_files(self):
        folder_path = os.path.join(test_dir, commit_dependency["Path"])
        os.makedirs(folder_path)
        file_path = os.path.join(folder_path, "test.txt")
        out_file = open(file_path, "w+")
        out_file.write("Make sure we don't delete this")
        out_file.close()
        self.assertTrue(os.path.exists(file_path))
        try:
            repo_resolver.resolve(test_dir, commit_dependency, force=True)
        except:
            self.fail("We shouldn't fail when we are forcing")
        details = repo_resolver.get_details(folder_path)
        self.assertEqual(details['Url'], commit_dependency['Url'])

    def test_wont_delete_dirty_repo(self):
        repo_resolver.resolve(test_dir, commit_dependency)

        folder_path = os.path.join(test_dir, commit_dependency["Path"])
        file_path = get_first_file(folder_path)
        # make sure the file already exists
        self.assertTrue(os.path.isfile(file_path))
        out_file = open(file_path, "a+")
        out_file.write("Make sure we don't delete this")
        out_file.close()
        self.assertTrue(os.path.exists(file_path))

        with self.assertRaises(Exception):
            repo_resolver.resolve(test_dir, commit_dependency, update_ok=True)

    def test_will_delete_dirty_repo(self):
        repo_resolver.resolve(test_dir, commit_dependency)
        folder_path = os.path.join(test_dir, commit_dependency["Path"])
        file_path = get_first_file(folder_path)
        # make sure the file already exists
        self.assertTrue(os.path.isfile(file_path))
        out_file = open(file_path, "a+")
        out_file.write("Make sure we don't delete this")
        out_file.close()
        self.assertTrue(os.path.exists(file_path))

        try:
            repo_resolver.resolve(test_dir, commit_later_dependency, force=True)
        except:
            self.fail("We shouldn't fail when we are forcing")

    # check to make sure we can clone a commit correctly
    def test_clone_commit_repo(self):
        # create an empty directory- and set that as the workspace
        repo_resolver.resolve(test_dir, commit_dependency)
        folder_path = os.path.join(test_dir, commit_dependency["Path"])
        details = repo_resolver.get_details(folder_path)

        self.assertEqual(details['Url'], commit_dependency['Url'])
        self.assertEqual(details['Commit'], commit_dependency['Commit'])

    # check to make sure we can clone a commit correctly
    def test_fail_update(self):
        # create an empty directory- and set that as the workspace
        repo_resolver.resolve(test_dir, commit_dependency)
        folder_path = os.path.join(test_dir, commit_dependency["Path"])
        details = repo_resolver.get_details(folder_path)

        self.assertEqual(details['Url'], commit_dependency['Url'])
        self.assertEqual(details['Commit'], commit_dependency['Commit'])
        # first we checkout
        with self.assertRaises(Exception):
            repo_resolver.resolve(test_dir, commit_later_dependency)

        details = repo_resolver.get_details(folder_path)
        self.assertEqual(details['Url'], commit_dependency['Url'])
        self.assertEqual(details['Commit'], commit_dependency['Commit'])

    def test_does_update(self):
        # create an empty directory- and set that as the workspace
        logging.info(f"Resolving at {test_dir}")
        repo_resolver.resolve(test_dir, commit_dependency)
        folder_path = os.path.join(test_dir, commit_dependency["Path"])
        logging.info(f"Getting details at {folder_path}")
        details = repo_resolver.get_details(folder_path)

        self.assertEqual(details['Url'], commit_dependency['Url'])
        self.assertEqual(details['Commit'], commit_dependency['Commit'])
        # next we checkout and go to the later commit
        try:
            repo_resolver.resolve(
                test_dir, commit_later_dependency, update_ok=True)
        except:
            self.fail("We are not supposed to throw an exception")
        details = repo_resolver.get_details(folder_path)
        logging.info(f"Checking {folder_path} for current git commit")

        self.assertEqual(details['Url'], commit_later_dependency['Url'])
        self.assertEqual(details['Commit'], commit_later_dependency['Commit'])

    def test_cant_switch_urls(self):
        # create an empty directory- and set that as the workspace
        repo_resolver.resolve(test_dir, branch_dependency)
        folder_path = os.path.join(test_dir, branch_dependency["Path"])

        details = repo_resolver.get_details(folder_path)

        self.assertEqual(details['Url'], branch_dependency['Url'])
        # first we checkout
        with self.assertRaises(Exception):
            repo_resolver.resolve(test_dir, microsoft_branch_dependency)

        details = repo_resolver.get_details(folder_path)
        self.assertEqual(details['Url'], branch_dependency['Url'])

    def test_ignore(self):
        # create an empty directory- and set that as the workspace
        repo_resolver.resolve(test_dir, branch_dependency)
        folder_path = os.path.join(test_dir, branch_dependency["Path"])

        details = repo_resolver.get_details(folder_path)

        self.assertEqual(details['Url'], branch_dependency['Url'])
        # first we checkout

        repo_resolver.resolve(
            test_dir, microsoft_branch_dependency, ignore=True)

        details = repo_resolver.get_details(folder_path)
        self.assertEqual(details['Url'], branch_dependency['Url'])

    def test_will_switch_urls(self):
        # create an empty directory- and set that as the workspace
        repo_resolver.resolve(test_dir, branch_dependency)

        folder_path = os.path.join(test_dir, branch_dependency["Path"])

        details = repo_resolver.get_details(folder_path)

        self.assertEqual(details['Url'], branch_dependency['Url'])
        # first we checkout
        try:
            repo_resolver.resolve(
                test_dir, microsoft_branch_dependency, force=True)
        except:
            self.fail("We shouldn't fail when we are forcing")

        details = repo_resolver.get_details(folder_path)
        self.assertEqual(details['Url'], microsoft_branch_dependency['Url'])

    def test_will_switch_branches(self):
        repo_resolver.resolve(test_dir, branch_dependency)
        folder_path = os.path.join(test_dir, branch_dependency["Path"])
        repo_resolver.resolve(test_dir, sub_branch_dependency, force=True)
        details = repo_resolver.get_details(folder_path)
        self.assertEqual(details['Url'], branch_dependency['Url'])
        self.assertEqual(details['Branch'], sub_branch_dependency['Branch'])


if __name__ == '__main__':
    unittest.main()



# this is for creating a minimal uefi tree for testing
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent

"""Used for creating a minimal uefi tree for testing."""
import os
import tempfile
import json
import random


class uefi_tree:
    """A minimal UEFI tree.

    This class represents a minimal UEFI tree that you can setup.
    It is configurable and offers options to modify the tree for.

    Attributes:
        workspace (str): Path to a directory for the minimal tree

    Args:
        workspace (str): Path to a directory for the minimal tree
        create_platform (bool): create the tree or use an existing
    """

    def __init__(self, workspace=None, create_platform=True):
        """Inits uefi_tree."""
        if workspace is None:
            workspace = os.path.abspath(tempfile.mkdtemp())
        self.workspace = workspace
        if (create_platform):
            self._create_tree()

    @staticmethod
    def _get_src_folder():
        return os.path.dirname(os.path.abspath(__file__))

    @staticmethod
    def _get_optional_folder():
        return os.path.join(uefi_tree._get_src_folder(), "optional")

    def get_workspace(self):
        """Returns the workspace path."""
        return self.workspace

    def _create_tree(self):
        """Creates a settings.py, test.dsc, Conf folder (with build_rule, target, and tools_def)."""
        settings_path = self.get_settings_provider_path()
        uefi_tree.write_to_file(settings_path, self._settings_file_text)
        dsc_path = os.path.join(self.workspace, "Test.dsc")
        uefi_tree.write_to_file(dsc_path, self._dsc_file_text)
        conf_path = os.path.join(self.workspace, "Conf")
        os.makedirs(conf_path, exist_ok=True)
        target_path = os.path.join(conf_path, "target.template")
        uefi_tree.write_to_file(target_path, self._target_file_text)
        build_rule_path = os.path.join(conf_path, "build_rule.template")
        uefi_tree.write_to_file(build_rule_path, "hello there")
        tools_path = os.path.join(conf_path, "tools_def.template")
        uefi_tree.write_to_file(tools_path, "hello there")

    @staticmethod
    def write_to_file(path, contents, close=True):
        """Writes contents to a file."""
        f = open(path, "w")
        f.writelines(contents)
        if close:
            f.close()

    def get_settings_provider_path(self):
        """Gets the settings provider in the workspace."""
        return os.path.join(self.workspace, "settings.py")

    def get_optional_file(self, file):
        """Gets the path of an optional file."""
        optional_path = os.path.join(uefi_tree._get_optional_folder(), file)
        if os.path.exists(optional_path):
            return os.path.abspath(optional_path)
        raise ValueError(f"Optional file not found {file}")

    def create_path_env(self, id=None, flags=[], var_name=None, scope="global", dir_path="", extra_data=None):
        """Creates an ext dep in your workspace."""
        data = {
            "scope": scope,
            "flags": flags,
        }
        if id is not None:
            data["id"] = id
        if var_name is not None:
            data["var_name"] = var_name
        if extra_data is not None:
            data.update(extra_data)
        text = json.dumps(data)
        if id is not None:
            file_name = f"{id}_path_env.json"
        else:
            file_name = "None%s_path_env.json" % str(random.randint(1, 100000))
        output_dir = os.path.join(self.workspace, dir_path)
        os.makedirs(output_dir, exist_ok=True)
        output_path = os.path.join(output_dir, file_name)
        uefi_tree.write_to_file(output_path, text)

    def create_Edk2TestUpdate_ext_dep(self, version="0.0.1", extra_data=None):
        """Creates an Edk2TestUpdate ext dep in your workspace."""
        self.create_ext_dep("nuget", "Edk2TestUpdate", version, extra_data=extra_data)

    def create_ext_dep(self, dep_type, name, version, source=None, scope="global", dir_path="", extra_data=None):
        """Creates an ext dep in your workspace."""
        dep_type = dep_type.lower()
        if source is None and dep_type == "nuget":
            source = "https://api.nuget.org/v3/index.json"
        if source is None:
            raise ValueError("Source was not provided")
        data = {
            "scope": scope,
            "type": dep_type,
            "name": name,
            "version": version,
            "source": source,
            "flags": []
        }
        if extra_data is not None:
            data.update(extra_data)
        text = json.dumps(data)
        raw_file_name = data["id"] if "id" in data else name
        ext_dep_name = raw_file_name.replace(" ", "_")
        file_name = f"{ext_dep_name}_ext_dep.json"
        output_dir = os.path.join(self.workspace, dir_path)
        os.makedirs(output_dir, exist_ok=True)
        output_path = os.path.join(output_dir, file_name)
        uefi_tree.write_to_file(output_path, text)

    _settings_file_text = '''
# @file settings.py
# This contains a settingsmanger for testing
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
from edk2toolext.environment.uefi_build import UefiBuilder
import os
from edk2toolext.invocables.edk2_platform_build import BuildSettingsManager
from edk2toolext.invocables.edk2_setup import SetupSettingsManager
from edk2toolext.invocables.edk2_ci_setup import Edk2CiBuildSetup
from edk2toolext.invocables.edk2_ci_setup import CiSetupSettingsManager
from edk2toolext.invocables.edk2_ci_build import CiBuildSettingsManager
from edk2toolext.invocables.edk2_update import UpdateSettingsManager


class TestSettingsManager(BuildSettingsManager, SetupSettingsManager, Edk2CiBuildSetup,
                          CiSetupSettingsManager, CiBuildSettingsManager, UpdateSettingsManager):

    def GetActiveScopes(self):
        return []

    def GetWorkspaceRoot(self):
        return os.path.dirname(__file__)

    def GetPackagesPath(self):
        return []

    def GetRequiredSubmodules(self):
        return []

    def GetDependencies(self):
        return []

    def AddCommandLineOptions(self, parserObj):
        pass

    def RetrieveCommandLineOptions(self, args):
        pass

    def GetName(self):
        return "TestEdk2Invocable"

    def GetArchitecturesSupported(self):
        return []

    def GetPackagesSupported(self):
        return []

    def GetTargetsSupported(self):
        return []

    def GetSkippedDirectories(self):
        return ()

class TestBuilder(UefiBuilder):
    def SetPlatformEnv(self):
        self.env.SetValue("EDK2_BASE_TOOLS_DIR", self.ws, "empty")
        return 0
    '''

    _dsc_file_text = '''
[Defines]
OUTPUT_DIRECTORY = Build
    '''

    _target_file_text = '''
ACTIVE_PLATFORM = Test.dsc
TOOL_CHAIN_TAG = test
TARGET_ARCH = X64
TARGET = DEBUG
    '''



# @file edk2_setup
# updates submodules listed as Required Submodules in Config file.
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Code that updates required submodules.

Contains a SetupSettingsManager that must be subclassed in a build settings
file. This provides platform specific information to Edk2PlatformSetup invocable
while allowing the invocable itself to remain platform agnostic.
"""
import os
import logging
from io import StringIO
from typing import List
from edk2toolext import edk2_logging
from edk2toolext.environment import version_aggregator
from edk2toolext.invocables.edk2_multipkg_aware_invocable import Edk2MultiPkgAwareInvocable
from edk2toolext.invocables.edk2_multipkg_aware_invocable import MultiPkgAwareSettingsInterface
from edk2toollib.utility_functions import RunCmd
from edk2toollib.utility_functions import version_compare


class RequiredSubmodule():
    """A class containing information about a git submodule."""
    def __init__(self, path: str, recursive: bool = True):
        """Object to hold necessary information for resolving submodules.

        Args:
            path (str): workspace relative path to submodule that must be
                synchronized and updated
            recursive (bool): if recursion should be used in this submodule
        """
        self.path = path
        self.recursive = recursive


class SetupSettingsManager(MultiPkgAwareSettingsInterface):
    """Platform specific settings for Edk2PlatformSetup.

    Provides information necessary for `stuart_setup.exe` or `edk2_setup.py`
    to successfully execute for a given platform.
    Example: Example: Overriding SetupSettingsManager
        ```python
        from edk2toolext.invocables.edk2_setup import SetupSettingsManager, RequiredSubmodule
        class PlatformManager(SetupSettingsManager):
            def GetRequiredSubmodules(self) -> List[RequiredSubmodule]:
                return [RequiredSubmodule('Common/MU', True)]
        ```
    """

    def GetRequiredSubmodules(self) -> List[RequiredSubmodule]:
        """Provides a list of required git submodules.

        These submodules are those that must be setup for the platform
        to successfully build.

         TIP: Optional Override in a subclass

        Returns:
            A list of required submodules, or an empty list
        """
        return []


class Edk2PlatformSetup(Edk2MultiPkgAwareInvocable):
    """Invocable that updates git submodules listed in RequiredSubmodules."""

    def AddCommandLineOptions(self, parserObj):
        """Adds command line options to the argparser."""
        parserObj.add_argument('--force', '--FORCE', '--Force', dest="force", action='store_true', default=False)
        parserObj.add_argument('--omnicache', '--OMNICACHE', '--Omnicache', dest='omnicache_path',
                               default=os.environ.get('OMNICACHE_PATH'))

        super().AddCommandLineOptions(parserObj)

    def RetrieveCommandLineOptions(self, args):
        """Retrieve command line options from the argparser."""
        self.force_it = args.force
        self.omnicache_path = args.omnicache_path
        if (self.omnicache_path is not None) and (not os.path.exists(self.omnicache_path)):
            logging.warning(f"Omnicache path set to invalid path: {args.omnicache_path}")
            self.omnicache_path = None

        super().RetrieveCommandLineOptions(args)

    def GetVerifyCheckRequired(self):
        """Will not call self_describing_environment.VerifyEnvironment because it hasn't been set up yet."""
        return False

    def GetSettingsClass(self):
        """Returns the SetupSettingsManager class.

        WARNING: SetupSettingsManager must be subclassed in your platform settings file.
        """
        return SetupSettingsManager

    def GetLoggingFileName(self, loggerType):
        """Returns the filename (SETUPLOG) of where the logs for the Edk2CiBuild invocable are stored in."""
        return "SETUPLOG"

    def Go(self):
        """Executes the core functionality of the Edk2PlatformSetup invocable."""
        required_submodules = self.PlatformSettings.GetRequiredSubmodules()
        workspace_path = self.GetWorkspaceRoot()
        # Make sure git is installed
        return_buffer = StringIO()
        RunCmd("git", "--version", outstream=return_buffer, raise_exception_on_nonzero=True)
        git_version = return_buffer.getvalue().strip()
        return_buffer.close()
        version_aggregator.GetVersionAggregator().ReportVersion("Git",
                                                                git_version,
                                                                version_aggregator.VersionTypes.TOOL)
        min_git = "2.11.0"
        # This code is highly specific to the return value of "git version"...
        cur_git = ".".join(git_version.split(' ')[2].split(".")[:3])
        if version_compare(min_git, cur_git) > 0:
            raise RuntimeError("Please upgrade Git! Current version is %s. Minimum is %s." % (cur_git, min_git))

        # Pre-setup cleaning if "--force" is specified.
        if self.force_it:
            try:
                # Clean and reset the main repo.
                edk2_logging.log_progress("## Cleaning the root repo...")
                RunCmd("git", "reset --hard", workingdir=workspace_path,
                       logging_level=logging.DEBUG, raise_exception_on_nonzero=True)
                # Because logging is running right now, we have to skip the files that are open.
                ignore_files = "-e Build/%s.txt -e Build/%s.md" % (self.GetLoggingFileName('txt'),
                                                                   self.GetLoggingFileName('md'))
                RunCmd("git", "clean -xffd %s" % ignore_files, workingdir=workspace_path,
                       logging_level=logging.DEBUG, raise_exception_on_nonzero=True)
                edk2_logging.log_progress("Done.\n")

                # Clean any submodule repos.
                if required_submodules:
                    for required_submodule in required_submodules:
                        edk2_logging.log_progress("## Cleaning Git repository: %s..." % required_submodule.path)
                        required_submodule_path = os.path.normpath(
                            os.path.join(workspace_path, required_submodule.path))
                        RunCmd("git", "reset --hard", workingdir=required_submodule_path,
                               logging_level=logging.DEBUG, raise_exception_on_nonzero=True)
                        RunCmd("git", "clean -xffd", workingdir=required_submodule_path,
                               logging_level=logging.DEBUG, raise_exception_on_nonzero=True)

                        edk2_logging.log_progress("Done.\n")

            except RuntimeError as e:
                logging.error("FAILED!\n")
                logging.error("Error while trying to clean the environment!")
                logging.error(str(e))
                return -1

        # Grab the remaining Git repos.
        result = 0
        if required_submodules and len(required_submodules) > 0:

            # Git Repos: STEP 1 --------------------------------------
            # Make sure that the repos are all synced.
            try:
                submodule_string = " ".join([x.path for x in required_submodules])
                edk2_logging.log_progress(f"## Syncing Git repositories: {submodule_string}...")
                RunCmd("git", f'submodule sync -- {submodule_string}',
                       workingdir=workspace_path, logging_level=logging.DEBUG, raise_exception_on_nonzero=True)

                edk2_logging.log_progress("Done.\n")
            except RuntimeError as e:
                logging.error("FAILED!\n")
                logging.error("Error while trying to synchronize the environment!")
                logging.error(str(e))
                return -1

            # Git Repos: STEP 2 --------------------------------------
            # Iterate through all repos and see whether they should be fetched.
            for required_submodule in required_submodules:
                try:
                    edk2_logging.log_progress(f"## Checking Git repository: {required_submodule.path}...")

                    # Git Repos: STEP 2a ---------------------------------
                    # Need to determine whether to skip this repo.
                    required_submodule_path = os.path.normpath(os.path.join(workspace_path, required_submodule.path))
                    skip_repo = False
                    # If the repo exists (and we're not forcing things) make
                    # sure that it's not in a "dirty" state.
                    if os.path.exists(required_submodule_path) and not self.force_it:
                        return_buffer = StringIO()
                        RunCmd("git", 'diff ' + required_submodule.path, outstream=return_buffer,
                               workingdir=workspace_path, logging_level=logging.DEBUG, raise_exception_on_nonzero=True)
                        git_data = return_buffer.getvalue().strip()
                        return_buffer.close()
                        # If anything was returned, we should skip processing the repo.
                        # It is either on a different commit or it has local changes.
                        if git_data != "":
                            logging.info("-- NOTE: Repo currently exists and appears to have local changes!")
                            logging.info("-- Skipping fetch!")
                            skip_repo = True

                    # Git Repos: STEP 2b ---------------------------------
                    # If we're not skipping, grab it.
                    if not skip_repo or self.force_it:
                        logging.info("## Fetching repo.")
                        cmd_string = "submodule update --init"
                        if required_submodule.recursive:
                            cmd_string += " --recursive"
                        cmd_string += " --progress"
                        if self.omnicache_path is not None:
                            cmd_string += " --reference " + self.omnicache_path
                        cmd_string += " " + required_submodule.path
                        ret = RunCmd('git', cmd_string, workingdir=workspace_path,
                                     logging_level=logging.DEBUG, raise_exception_on_nonzero=False)
                        if ret != 0:
                            logging.error("Failed to fetch " + required_submodule.path)
                            raise RuntimeError("Unable to checkout repo due to error")

                    edk2_logging.log_progress("Done.\n")

                except RuntimeError as e:
                    logging.error("FAILED!\n")
                    logging.error("Failed to fetch required repository!\n")
                    logging.error(str(e))
                    result = -1

        return result


def main():
    """Entry point to invoke Edk2PlatformSetup."""
    Edk2PlatformSetup().Invoke()



# @file test_pyopenssl_signer.py
# This contains unit tests for the pyopenssl binary wrapper
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import unittest
from edk2toolext.capsule import pyopenssl_signer
from OpenSSL import crypto


class Test_pyopenssl_signer(unittest.TestCase):

    def test_empty(self):
        with self.assertRaises((KeyError, ValueError)):
            pyopenssl_signer.sign(None, {}, {})

    def test_proper_options(self):
        # we're going to assume that we're
        with self.assertRaises(crypto.Error):
            signer = {
                'key_file_format': 'pkcs12',
                'key_data': "hello there"
            }
            signature = {
                'type': 'bare',
                'encoding': 'binary',
                'hash_alg': 'sha256',

            }
            pyopenssl_signer.sign(None, signature, signer)

    def test_invalid_type(self):
        # we're going to assume that we're
        with self.assertRaises(ValueError):
            signature = {
                'type': 'bad_type',
            }
            pyopenssl_signer.sign(None, signature, {})

    def test_invalid_type_options(self):
        # we're going to assume that we're
        with self.assertRaises(ValueError):
            signature = {
                'type': 'bare',
                'type_options': 'not allowed'
            }
            pyopenssl_signer.sign(None, signature, {})



# @file test_git_dependency.py
# Unit test suite for the GitDependency class.
#
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import os
import sys
import unittest
import logging
import tempfile
import pkg_resources
from edk2toollib.utility_functions import RunCmd, RemoveTree
from edk2toolext.environment import environment_descriptor_files as EDF
from edk2toolext.environment.extdeptypes.nuget_dependency import NugetDependency
from edk2toolext.environment import version_aggregator
from edk2toolext.bin import nuget

test_dir = None
good_version = "5.2.0"
bad_version = "5.2.13.1.2"
missing_version = "5.200.13"

hw_package_name = "NuGet.CommandLine"
hw_json_template = '''
{
  "scope": "global",
  "type": "nuget",
  "name": "NuGet.CommandLine",
  "source": "https://api.nuget.org/v3/index.json",
  "version": "%s"
}
'''


def prep_workspace():
    global test_dir
    # if test temp dir doesn't exist
    if test_dir is None or not os.path.isdir(test_dir):
        test_dir = tempfile.mkdtemp()
        logging.debug("temp dir is: %s" % test_dir)
    else:
        clean_workspace()
        test_dir = tempfile.mkdtemp()


def clean_workspace():
    global test_dir
    if test_dir is None:
        return

    if os.path.isdir(test_dir):
        RemoveTree(test_dir)
        test_dir = None


class TestNugetDependency(unittest.TestCase):
    def setUp(self):
        prep_workspace()
        self.saved_nuget_path = os.getenv(NugetDependency.NUGET_ENV_VAR_NAME)

    @classmethod
    def setUpClass(cls):
        logger = logging.getLogger('')
        logger.addHandler(logging.NullHandler())
        unittest.installHandler()

    @classmethod
    def tearDownClass(cls):
        clean_workspace()

    def tearDown(self):
        # we need to reset the version aggregator each time
        version_aggregator.GetVersionAggregator().Reset()

        if self.saved_nuget_path is not None:
            os.environ[NugetDependency.NUGET_ENV_VAR_NAME] = self.saved_nuget_path

        # Fix the nuget.exe is missing....download again
        requirement = pkg_resources.Requirement.parse("edk2-pytool-extensions")
        nuget_file_path = os.path.join("edk2toolext", "bin", "NuGet.exe")
        nuget_path = pkg_resources.resource_filename(requirement, nuget_file_path)

        if not os.path.isfile(nuget_path):
            nuget.DownloadNuget()

    def test_can_get_nuget_path(self):
        nuget_cmd = NugetDependency.GetNugetCmd()
        nuget_cmd += ["locals", "global-packages", "-list"]
        ret = RunCmd(nuget_cmd[0], ' '.join(nuget_cmd[1:]), outstream=sys.stdout)
        self.assertEqual(ret, 0)  # make sure we have a zero return code

    def test_missing_nuget(self):

        if NugetDependency.NUGET_ENV_VAR_NAME in os.environ:
            del os.environ[NugetDependency.NUGET_ENV_VAR_NAME]

        # delete the package file
        original = NugetDependency.GetNugetCmd()[-1]  # get last item which will be exe path
        os.remove(original)
        path = NugetDependency.GetNugetCmd()
        self.assertIsNone(path)  # Should not be found

    def test_nuget_env_var(self):
        if NugetDependency.NUGET_ENV_VAR_NAME in os.environ:
            del os.environ[NugetDependency.NUGET_ENV_VAR_NAME]

        # set the env var to our path
        os.environ[NugetDependency.NUGET_ENV_VAR_NAME] = test_dir
        nuget.DownloadNuget(test_dir)  # download to test dir
        found_path = NugetDependency.GetNugetCmd()[-1]

        # done with env testing.  clean up
        del os.environ[NugetDependency.NUGET_ENV_VAR_NAME]
        self.assertIsNotNone(found_path)
        path_should_be = os.path.join(test_dir, "NuGet.exe")
        self.assertTrue(os.path.samefile(found_path, path_should_be))

    def test_nuget_env_var_with_space(self):
        if NugetDependency.NUGET_ENV_VAR_NAME in os.environ:
            del os.environ[NugetDependency.NUGET_ENV_VAR_NAME]

        # set the env var to our path
        my_test_dir = os.path.join(test_dir, "my folder")
        os.makedirs(my_test_dir)
        os.environ[NugetDependency.NUGET_ENV_VAR_NAME] = my_test_dir
        nuget.DownloadNuget(my_test_dir)  # download to test dir
        found_path = NugetDependency.GetNugetCmd()[-1]

        # done with env testing.  clean up
        del os.environ[NugetDependency.NUGET_ENV_VAR_NAME]

        self.assertIsNotNone(found_path)
        path_should_be = os.path.join(my_test_dir, "NuGet.exe")
        self.assertTrue(os.path.samefile(found_path.strip('"'), path_should_be))

    # good case
    def test_download_good_nuget(self):
        ext_dep_file_path = os.path.join(test_dir, "hw_ext_dep.json")
        with open(ext_dep_file_path, "w+") as ext_dep_file:
            ext_dep_file.write(hw_json_template % good_version)

        ext_dep_descriptor = EDF.ExternDepDescriptor(ext_dep_file_path).descriptor_contents
        ext_dep = NugetDependency(ext_dep_descriptor)
        ext_dep.fetch()
        self.assertTrue(ext_dep.verify())
        self.assertEqual(ext_dep.version, good_version)
        # make sure we clean up after ourselves
        ext_dep.clean()

    # bad case
    def test_download_bad_nuget(self):
        ext_dep_file_path = os.path.join(test_dir, "hw_ext_dep.json")
        with open(ext_dep_file_path, "w+") as ext_dep_file:
            ext_dep_file.write(hw_json_template % bad_version)

        ext_dep_descriptor = EDF.ExternDepDescriptor(ext_dep_file_path).descriptor_contents
        ext_dep = NugetDependency(ext_dep_descriptor)
        with self.assertRaises((RuntimeError, ValueError)):  # we can throw a value error if we hit the cache
            # we should throw an exception because we don't know how to parse the version
            ext_dep.fetch()
        self.assertFalse(ext_dep.verify())

    def test_normalize_version(self):
        version = "5.10.05.0"
        proper_version = "5.10.5"
        self.assertEqual(proper_version, NugetDependency.normalize_version(version))
        version = "6.10"
        proper_version = "6.10.0"
        self.assertEqual(proper_version, NugetDependency.normalize_version(version))
        version = "6"
        proper_version = "6.0.0"
        self.assertEqual(proper_version, NugetDependency.normalize_version(version))
        version = "6-beta"
        proper_version = "6.0.0-beta"
        self.assertEqual(proper_version, NugetDependency.normalize_version(version))
        version = "3.2.1.-alpha"
        proper_version = "3.2.1-alpha"
        self.assertEqual(proper_version, NugetDependency.normalize_version(version))
        version = "3.2.1.0-rc1"
        proper_version = "3.2.1-rc1"
        self.assertEqual(proper_version, NugetDependency.normalize_version(version))
        version = "1.18.9.2"
        proper_version = "1.18.9.2"
        self.assertEqual(proper_version, NugetDependency.normalize_version(version))
        version = "0.4.0"
        proper_version = "0.4.0"
        self.assertEqual(proper_version, NugetDependency.normalize_version(version))
        version = "21.0.1056.1"
        proper_version = "21.0.1056.1"
        self.assertEqual(proper_version, NugetDependency.normalize_version(version))
        version = "2022.02.1"
        proper_version = "2022.2.1"
        self.assertEqual(proper_version, NugetDependency.normalize_version(version))
        version = "2.15.05"
        proper_version = "2.15.5"
        self.assertEqual(proper_version, NugetDependency.normalize_version(version))
        version = "2.8.4-92"
        proper_version = "2.8.4-92"
        self.assertEqual(proper_version, NugetDependency.normalize_version(version))

        # try some bad cases
        with self.assertRaises(ValueError):
            NugetDependency.normalize_version("not a number")
        with self.assertRaises(ValueError):
            NugetDependency.normalize_version("6.0-beta-beta")
        with self.assertRaises(ValueError):
            NugetDependency.normalize_version("6.0-")
        with self.assertRaises(ValueError):
            NugetDependency.normalize_version("--")
        with self.assertRaises(ValueError):
            NugetDependency.normalize_version("6-")
        with self.assertRaises(ValueError):
            NugetDependency.normalize_version("")
        with self.assertRaises(ValueError):
            NugetDependency.normalize_version(bad_version)

    # missing case
    def test_download_missing_nuget(self):
        ext_dep_file_path = os.path.join(test_dir, "hw_ext_dep.json")
        with open(ext_dep_file_path, "w+") as ext_dep_file:
            ext_dep_file.write(hw_json_template % missing_version)

        ext_dep_descriptor = EDF.ExternDepDescriptor(ext_dep_file_path).descriptor_contents
        ext_dep = NugetDependency(ext_dep_descriptor)
        with self.assertRaises(RuntimeError):
            ext_dep.fetch()
        self.assertFalse(ext_dep.verify())
        self.assertEqual(ext_dep.version, missing_version)

    def test_cache_path_not_found(self):
        ext_dep_file_path = os.path.join(test_dir, "hw_ext_dep.json")
        with open(ext_dep_file_path, "w+") as ext_dep_file:
            ext_dep_file.write(hw_json_template % good_version)

        ext_dep_descriptor = EDF.ExternDepDescriptor(ext_dep_file_path).descriptor_contents
        ext_dep = NugetDependency(ext_dep_descriptor)

        ext_dep.nuget_cache_path = "not_a_real_path"
        self.assertFalse(ext_dep._fetch_from_nuget_cache(hw_package_name))

    def test_bad_cached_package(self):
        ext_dep_file_path = os.path.join(test_dir, "hw_ext_dep.json")
        with open(ext_dep_file_path, "w+") as ext_dep_file:
            ext_dep_file.write(hw_json_template % good_version)

        ext_dep_descriptor = EDF.ExternDepDescriptor(ext_dep_file_path).descriptor_contents
        ext_dep = NugetDependency(ext_dep_descriptor)

        #
        # Create a cache with a bad cached package.
        #
        # First, create the cache.
        cache_dir = os.path.join(test_dir, 'nuget_test_bad_cache')
        os.mkdir(cache_dir)
        ext_dep.nuget_cache_path = cache_dir
        # Then create the directories inside the cache that should hold the contents.
        package_cache_dir = os.path.join(cache_dir, hw_package_name.lower(), good_version)
        os.makedirs(package_cache_dir)
        # There are no package directories inside the cache.
        self.assertFalse(ext_dep._fetch_from_nuget_cache(hw_package_name))

        # Create a directory that doesn't match the heuristic.
        test_cache_contents = os.path.join(package_cache_dir, "contents", "blah")
        os.makedirs(test_cache_contents)
        self.assertFalse(ext_dep._fetch_from_nuget_cache(hw_package_name))

    def test_good_cached_package(self):
        ext_dep_file_path = os.path.join(test_dir, "hw_ext_dep.json")
        with open(ext_dep_file_path, "w+") as ext_dep_file:
            ext_dep_file.write(hw_json_template % good_version)

        ext_dep_descriptor = EDF.ExternDepDescriptor(ext_dep_file_path).descriptor_contents
        ext_dep = NugetDependency(ext_dep_descriptor)

        #
        # Create a cache with a good cached package.
        #
        # First, create the cache.
        cache_dir = os.path.join(test_dir, 'nuget_test_good_cache')
        os.mkdir(cache_dir)
        ext_dep.nuget_cache_path = cache_dir
        # Then create the directories inside the cache that should hold the contents.
        package_cache_dir = os.path.join(cache_dir, hw_package_name.lower(), good_version)
        os.makedirs(package_cache_dir)

        # Create a directory that does match the heuristic.
        test_cache_contents = os.path.join(package_cache_dir, hw_package_name, "working_blah")
        os.makedirs(test_cache_contents)
        self.assertTrue(ext_dep._fetch_from_nuget_cache(hw_package_name))
        ext_dep.update_state_file()

        # Make sure that the contents were copied correctly.
        final_path = os.path.join(ext_dep.contents_dir, "working_blah")
        self.assertTrue(os.path.isdir(final_path))
        self.assertTrue(ext_dep.verify())


if __name__ == '__main__':
    unittest.main()



# @file
# Basic command-line interface for creating
# and decoding Windows Firmware Policy blobs
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""A basic command-line interface for creating and decoding windows firmware policy blobs."""

from edk2toollib.windows.policy.firmware_policy import FirmwarePolicy
import argparse


def PrintPolicy(filename: str) -> None:
    """Attempts to parse filename as a Windows Firmware Policy and print it."""
    try:
        with open(filename, 'rb') as f:
            policy = FirmwarePolicy(fs=f)
            policy.Print()

    except FileNotFoundError:
        print('ERROR:  File not found: "{0}"'.format(filename))


def CreatePolicyFromParameters(filename: str, manufacturer: str, product: str,
                               sn: str, nonce: int, oem1: str, oem2: str, devicePolicy: int) -> None:
    """Populates a Windows FirmwarePolicy object with the provided parameters and serializes it to filename.

    WARNING: Filename must be a new file, will not overwrite existing files.
    """
    with open(filename, 'xb') as f:
        policy = FirmwarePolicy()
        TargetInfo = {'Manufacturer': manufacturer,
                      'Product': product,
                      'SerialNumber': sn,
                      'OEM_01': oem1,
                      'OEM_02': oem2,
                      'Nonce': nonce}
        policy.SetDeviceTarget(TargetInfo)
        policy.SetDevicePolicy(devicePolicy)
        policy.SerializeToStream(stream=f)
        policy.Print()


def main():
    """Parses command-line parameters using ArgumentParser, passing them to helper functions to perform the requests."""
    parser = argparse.ArgumentParser(description='Firmware Policy Tool')
    subparsers = parser.add_subparsers(required=True, dest='action')

    parser_create = subparsers.add_parser('create', help='Create a firmware policy')
    parser_create.add_argument('PolicyFilename', type=str, help='The name of the new binary policy file to create '
                               '- will not overwrite existing files')
    parser_create.add_argument(
        'Manufacturer', type=str, help='Manufacturer Name, for example, "Contoso Computers, LLC".  '
        'Should match the EV Certificate Subject CN="Manufacturer"')
    parser_create.add_argument('Product', type=str, help='Product Name, for example, "Laptop Foo"')
    parser_create.add_argument(
        'SerialNumber', type=str, help='Serial Number, for example "F0013-000243546-X02".  Should match '
        'SmbiosSystemSerialNumber, SMBIOS System Information (Type 1 Table) -> Serial Number')
    parser_create.add_argument('NonceHex', type=str, help='The nonce in hexadecimal, for example "0x0123456789abcdef"')
    parser_create.add_argument('--OEM1', type=str, default='',
                               help='Optional OEM Field 1, an arbitrary length string, for example "ODM foo"')
    parser_create.add_argument('--OEM2', type=str, default='', help='Optional OEM Field 2, an arbitrary length string')
    parser_create.add_argument('DevicePolicyHex', type=str, help='The device policy in hexadecimal,'
                               ' for example to clear the TPM and delete Secure Boot keys: 0x3')

    parser_print = subparsers.add_parser('parse', help='Parse a firmware policy and print in human readable form')
    parser_print.add_argument('filename', help='Filename to parse and print')

    options = parser.parse_args()

    print('Options: ', options)

    if options.action == 'create':
        nonceInt = int(options.NonceHex, 16)
        devicePolicy = int(options.DevicePolicyHex, 16)
        CreatePolicyFromParameters(options.PolicyFilename, options.Manufacturer,
                                   options.Product, options.SerialNumber, nonceInt,
                                   options.OEM1, options.OEM2, devicePolicy)

    elif options.action == 'parse':
        PrintPolicy(options.filename)


if __name__ == '__main__':
    main()



# @file UefiBuildPlugin
# Plugin that supports Pre and Post Build steps
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Plugin that supports Pre and Post Build Steps."""


class IUefiBuildPlugin(object):
    """Plugin that supports Pre and Post Build Steps."""

    def do_post_build(self, thebuilder):
        """Runs Post Build Plugin Operations.

        Args:
            thebuilder (UefiBuilder): UefiBuild object for env information

        Returns:
            (int): 0 or NonZero for success or failure
        """
        return 0

    def do_pre_build(self, thebuilder):
        """Runs Pre Build Plugin Operations.

        Args:
            thebuilder (UefiBuilder): UefiBuild object for env information

        Returns:
            (int): 0 or NonZero for success or failure
        """
        return 0



# @file test_self_describing_environment.py
# This contains unit tests for the SDE
#
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import os
import unittest
import tempfile
from edk2toolext.environment import self_describing_environment
from edk2toolext.tests.uefi_tree import uefi_tree
from edk2toolext.environment import version_aggregator


class Testself_describing_environment(unittest.TestCase):

    def setUp(self):
        self.workspace = os.path.abspath(tempfile.mkdtemp())
        # we need to make sure to tear down the version aggregator and the SDE
        self_describing_environment.DestroyEnvironment()
        version_aggregator.ResetVersionAggregator()

    def test_null_init(self):
        sde = self_describing_environment.self_describing_environment(self.workspace)
        self.assertIsNotNone(sde)

    def test_unique_scopes_required(self):
        ''' make sure the sde will throw exception if duplicate scopes are specified '''
        scopes = ("corebuild", "corebuild", "testing", "CoreBuild")
        with self.assertRaises(ValueError):
            self_describing_environment.self_describing_environment(self.workspace, scopes)

    def test_collect_path_env(self):
        ''' makes sure the SDE can collect path env '''
        scopes = ("global",)
        tree = uefi_tree(self.workspace, create_platform=False)
        tree.create_path_env("testing_corebuild", var_name="hey", flags=["set_path", ])
        tree.create_path_env("testing_corebuild2", var_name="hey", flags=["set_pypath", ])
        tree.create_path_env("testing_corebuild3", var_name="hey", flags=["set_build_var", ])
        tree.create_path_env("testing_corebuild4", var_name="hey", flags=["set_shell_var", ])
        build_env, shell_env = self_describing_environment.BootstrapEnvironment(self.workspace, scopes)
        self.assertEqual(len(build_env.paths), 4)

    def test_collect_path_env_scoped(self):
        ''' makes sure the SDE can collect path env with the right scopes '''
        scopes = ("global", "testing")
        tree = uefi_tree(self.workspace, create_platform=False)
        tree.create_path_env("testing_corebuild", scope="testing")
        tree.create_path_env("testing_corebuild2", scope="not_valid")
        build_env, shell_env = self_describing_environment.BootstrapEnvironment(self.workspace, scopes)
        self.assertEqual(len(build_env.paths), 1)

    def test_override_path_env(self):
        ''' checks the SDE descriptor override system '''
        custom_scope = "global"
        scopes = (custom_scope,)
        tree = uefi_tree(self.workspace, create_platform=False)
        tree.create_path_env("testing_corebuild", var_name="hey", dir_path="test1", scope=custom_scope)
        tree.create_path_env("testing_corebuild2", var_name="jokes", scope=custom_scope,
                             extra_data={"override_id": "testing_corebuild"})
        build_env, shell_env = self_describing_environment.BootstrapEnvironment(self.workspace, scopes)
        self.assertEqual(len(build_env.paths), 1)

    def test_multiple_override_path_env(self):
        ''' checks the SDE descriptor override system will throw an error on multiple overrides'''
        custom_scope = "global"
        scopes = (custom_scope,)
        tree = uefi_tree(self.workspace, create_platform=False)
        tree.create_path_env("testing_corebuild", var_name="hey", dir_path="test1", scope=custom_scope)
        tree.create_path_env("testing_corebuild2", var_name="jokes", scope=custom_scope,
                             extra_data={"override_id": "testing_corebuild"})
        tree.create_path_env("testing_corebuild3", var_name="laughs", scope=custom_scope,
                             extra_data={"override_id": "testing_corebuild"})
        # we should get an exception because we have two overrides
        with self.assertRaises(RuntimeError):
            build_env, shell_env = self_describing_environment.BootstrapEnvironment(self.workspace, scopes)
            self.fail()

    def test_override_path_env_swapped_order(self):
        ''' checks the SDE descriptor override system with reversed paths so they are discovered in opposite order'''
        custom_scope = "global"
        scopes = (custom_scope,)
        tree = uefi_tree(self.workspace, create_platform=False)
        tree.create_path_env("testing_corebuild", var_name="hey", scope=custom_scope)
        tree.create_path_env(var_name="jokes", dir_path="test1", scope=custom_scope,
                             extra_data={"override_id": "testing_corebuild"})
        build_env, shell_env = self_describing_environment.BootstrapEnvironment(self.workspace, scopes)
        self.assertEqual(len(build_env.paths), 1)

    def test_duplicate_id_path_env(self):
        ''' check that the SDE will throw an exception if path_env have duplicate id's '''
        custom_scope = "global"
        scopes = (custom_scope,)
        tree = uefi_tree(self.workspace, create_platform=False)
        tree.create_path_env("testing_corebuild", dir_path="test1")
        tree.create_path_env("testing_corebuild")
        with self.assertRaises(RuntimeError):
            self_describing_environment.BootstrapEnvironment(self.workspace, scopes)
            self.fail()

    def test_duplicate_id_path_env_2(self):
        ''' check that the SDE will throw an exception if path env have duplicate id's.
        Since id is not a required member of path env make sure it can handle case where one of the path
        env files doesn't define an id'''
        custom_scope = "global"
        scopes = (custom_scope,)
        tree = uefi_tree(self.workspace, create_platform=False)
        tree.create_path_env("testing_corebuild", dir_path="test1")
        tree.create_path_env("testing_corebuild")
        tree.create_path_env()
        with self.assertRaises(RuntimeError):
            self_describing_environment.BootstrapEnvironment(self.workspace, scopes)
            self.fail()


if __name__ == '__main__':
    unittest.main()



# @file signtool_signer.py
# This module contains the abstracted signing interface for Windows Signtool.
# This interface abstraction takes in the signature_options and signer_options
# dictionaries that are used by capsule_tool and capsule_helper.
#
# Will attempt to locate a valid installation of Windows Signtool using the
# utility_functions provided by edk2toollib.
#
##
# Copyright (C) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Abstracted signing interface for Windows Signtool.

This interface abstraction takes in the signature_options and signer_options
dictionaries that are used by capsule_tool and capsule_helper.

Will attempt to locate a valid installation of Windows Signtool using the
utility_functions provided by edk2toollib.
"""
import os
import tempfile
import warnings

from edk2toollib.windows import locate_tools
from edk2toollib.utility_functions import RunCmd

GLOBAL_SIGNTOOL_PATH = None
SUPPORTED_SIGNATURE_TYPE_OPTIONS = {
    'pkcs7': {'detachedSignedData', 'embedded', 'pkcs7DetachedSignedData'}
}


def get_signtool_path():
    """Locates a valid installation of Windows Signtool.

    Will attempt to reuse a previously located version, since this call can be
    lengthy.
    """
    global GLOBAL_SIGNTOOL_PATH

    if GLOBAL_SIGNTOOL_PATH is None:
        GLOBAL_SIGNTOOL_PATH = locate_tools.FindToolInWinSdk('signtool.exe')

    return GLOBAL_SIGNTOOL_PATH


def sign(data: bytes, signature_options: dict, signer_options: dict) -> bytes:
    """Primary signing interface.

    Takes in the signature_options and signer_options
    dictionaries that are used by capsule_tool and capsule_helper

    Args:
        data (bytes): data to write into the sign tool.
        signature_options (dict): dictionary containing signature options
        signer_options (dict): dictionary containing signer options

    Raises:
        (ValueError()): Unsupported signature or signer options
        (RuntimeError()): Signtool.exe returned with error

    Note: Currently, only support the necessary options for capsules & Windows Firmware Policies
    """
    # The following _if_ clause handles the deprecated signature_option 'sign_alg' for backwards compatibility
    # when the deprecated option is supplied, this code adds the new, required options based on prior code behavior
    if 'sign_alg' in signature_options:
        warnings.warn('Signature_option "sign_alg" is deprecated, use "type"', DeprecationWarning)
        if signature_options['sign_alg'] == 'pkcs12':
            # map legacy behavior to new options and backwards-compatible values
            signature_options['type'] = 'pkcs7'
            signature_options['type_options'] = {'detachedSignedData'}
            signature_options['encoding'] = 'DER'
            signer_options['key_file_format'] = 'pkcs12'
        else:
            raise ValueError(f"Unsupported signature algorithm: {signature_options['sign_alg']}!")

    if signature_options['type'] != 'pkcs7':
        raise ValueError(f"Unsupported signature type: {signature_options['type']}!")
    for opt in signature_options['type_options']:
        if opt not in SUPPORTED_SIGNATURE_TYPE_OPTIONS[signature_options['type']]:
            raise ValueError(f"Unsupported type option: {opt}!  Ensure you have provied a set")

    mutually_exclusive_options = ('embedded', 'detachedSignedData', 'pkcs7DetachedSignedData')
    option_found = None
    for option in mutually_exclusive_options:
        if option in signature_options['type_options']:
            if option_found is None:
                option_found = option
            else:
                raise ValueError("type_options '%s' and '%s' are mutually exclusive" % (option_found, option))

    if signature_options['encoding'] != 'DER':
        raise ValueError(f"Unsupported signature encoding: {signature_options['type']}!")
    if signature_options['hash_alg'] != 'sha256':
        raise ValueError(f"Unsupported hashing algorithm: {signature_options['hash_alg']}!")
    if 'key_file' not in signer_options:
        raise ValueError("Must supply a key_file in signer_options for Signtool!")
    if signer_options['key_file_format'] != 'pkcs12':
        raise ValueError(f"Unsupported key file format: {signer_options['key_file_format']}!")

    # Set up a temp directory to hold input and output files.
    temp_folder = tempfile.mkdtemp()
    in_file_path = os.path.join(temp_folder, "data_to_sign.bin")

    # Create the input file for Signtool.
    in_file = open(in_file_path, 'wb')
    in_file.write(data)
    in_file.close()

    # Start building the parameters for the call.
    signtool_params = ['sign']
    signtool_params += ['/fd', signature_options['hash_alg']]
    if 'detachedSignedData' in signature_options['type_options']:
        signtool_params += ['/p7ce', 'DetachedSignedData']
    elif 'pkcs7DetachedSignedData' in signature_options['type_options']:
        signtool_params += ['/p7ce', 'PKCS7DetachedSignedData']
    elif 'embedded' in signature_options['type_options']:
        signtool_params += ['/p7ce', 'Embedded']
    else:
        raise ValueError("For pkcs7, type_options must include either embedded or detachedSignedData")
    signtool_params += ['/p7', f'"{temp_folder}"']
    signtool_params += ['/f', f"\"{signer_options['key_file']}\""]
    if 'oid' in signer_options:
        signtool_params += ['/p7co', signer_options['oid']]
    if 'eku' in signer_options:
        signtool_params += ['/u', signer_options['eku']]
    if 'key_pass' in signer_options:
        signtool_params += ['/p', signer_options['key_pass']]
    # Add basic options.
    signtool_params += ['/debug', '/v', f'"{in_file_path}"']

    # Make the call to Signtool.
    ret = RunCmd(get_signtool_path(), " ".join(signtool_params))
    if ret != 0:
        raise RuntimeError(f"Signtool.exe returned with error: {ret}!")

    # Load the data from the output file and return it.
    out_file_path = os.path.join(temp_folder, "data_to_sign.bin.p7")
    out_file = open(out_file_path, 'rb')
    out_data = out_file.read()
    out_file.close()

    return out_data


def sign_in_place(sign_file_path, signature_options, signer_options):
    """Alternate module-specific signing interface.

    Supports particular signatures associated with windows capsule files
    (e.g. CAT files). Takes in the signature_options and signer_options
    dictionaries that are used by capsule_tool and capsule_helper.

    Args:
        sign_file_path (str): sign file path
        signature_options (Dict): dictionary containing signature options
        signer_options (Dict): dictionary containing signer options

    Raises:
        (ValueError()): Unsupported signature or signer options
        (RuntimeError()): Signtool.exe returned with error.

    """
    # NOTE: Currently, we only support the necessary algorithms for capsules.
    if signature_options['sign_alg'] != 'pkcs12':
        raise ValueError(f"Unsupported signature algorithm: {signature_options['sign_alg']}!")
    if signature_options['hash_alg'] != 'sha256':
        raise ValueError(f"Unsupported hashing algorithm: {signature_options['hash_alg']}!")
    if 'key_file' not in signer_options:
        raise ValueError("Must supply a key_file in signer_options for Signtool!")

    # Start building the parameters for the call.
    signtool_params = ['sign', '/a']
    signtool_params += ['/fd', signature_options['hash_alg']]
    signtool_params += ['/f', f"\"{signer_options['key_file']}\""]
    # if 'oid' in signer_options:
    #     signtool_params += ['/p7co', signer_options['oid']]
    # if 'eku' in signer_options:
    #     signtool_params += ['/u', signer_options['eku']]
    if 'key_pass' in signer_options:
        signtool_params += ['/p', signer_options['key_pass']]
    # Add basic options.
    signtool_params += ['/debug', '/v', f'"{sign_file_path}"']

    # Make the call to Signtool.
    ret = RunCmd(get_signtool_path(), " ".join(signtool_params))
    if ret != 0:
        raise RuntimeError(f"Signtool.exe returned with error: {ret}!")



# @file edk2_multipkg_aware_invocable
# An intermediate class that supports a multi-package aware
# invocable process.
#
# Add cmdline parameter handling, a base settings manager class,
# and a Callback.
#
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""An intermediate class that supports a multi-package aware invocable process.

Provides two main classes, the MultiPkgAwareSettingsInterface and the
Edk2MultiPkgAwareInvocable that act as an intermediate class that other
invocables that require a multi-package aware invocable process. These classes
should only be subclassed if a new invocable is being developed. Any
Edk2MultiPkgAwareInvocable should be platform agnostic and work for any
platform. Platform specific data is provided via the
MultiPkgAwareSettingsInterface
"""
from edk2toolext.edk2_invocable import Edk2Invocable, Edk2InvocableSettingsInterface


class MultiPkgAwareSettingsInterface(Edk2InvocableSettingsInterface):
    """Settings to support Multi-Pkg functionality.

    This is an interface definition only to show which functions are required
    to be implemented and which functions can be implemented.

    Example: Overriding MultiPkgAwareSettingsInterface
        ``` python
        import os
        import logging
        import argparse
        from typing import Iterable, Tuple
        from edk2toolext.edk2_multipkg_aware_invocable import MultiPkgAwareSettingsInterface
        class NewInvocableSettingsManager(MultiPkgAwareSettingsInterface):

            def GetPackagesSupported(self):
                return ("PlatformPkg",)

            def GetArchitecturesSupported(self):
                return ("IA32","X64)

            def GetTargetsSupported(self):
                return ("TARGET", "RELEASE")

            def SetPackages(self, list_of_requested_packages):

                if len(filter(lambda pkg: pkg in self.GetPackagesSupported(), list_of_requested_packages)) !=
                   len(list_of_requested_packages):
                   raise Exception("Requested Packages contains unsupported Package")
                else:
                    self.pkgs = list_of_requested_packages

            def SetArchitectures(self, list_of_requested_architectures):
                if list_of_requested_architectures != self.GetPackagesSupported():
                    raise Exception("Only Support IA32,X64 combination")
            def SetTargets(self, list_of_requested_targets):
                if list_of_requested_targets != self.GetArchitecturesSupported():
                    raise Exception("Only Support "TARGET", "RELEASE combination")
        ```

    Warning:
        This interface should not be subclassed directly unless creating a new invocable type. Override these
        methods as a part of other subclasses invocable settings managers such as SetupSettingsManager, etc.
    """

    # ####################################################################################### #
    #                           Supported Values and Defaults                                 #
    # ####################################################################################### #
    def GetPackagesSupported(self):
        """Returns an iterable of edk2 packages supported by this build.

        TIP: Required Override in a subclass

        Returns:
            (Iterable): edk2 packages

        Note:
            packages should be relative to workspace or package path
        """
        raise NotImplementedError()

    def GetArchitecturesSupported(self):
        """Returns an iterable of edk2 architectures supported by this build.

        TIP: Required Override in a subclass

        Returns:
            (Iterable): architectures (X64, I32, etc.)
        """
        raise NotImplementedError()

    def GetTargetsSupported(self):
        """Returns an iterable of edk2 target tags supported by this build.

        TIP: Required Override in a subclass

        Returns:
            (Iterable): targets (DEBUG, RELEASE, etc)
        """
        raise NotImplementedError()

    # ####################################################################################### #
    #                     Verify and Save requested Config                                    #
    # ####################################################################################### #
    def SetPackages(self, list_of_requested_packages):
        """Confirms the requested package list is valid.

        TIP: Optional Override in a subclass

        Args:
            list_of_requested_packages (list[str]): packages to be built

        Raises:
            Exception: A requested package is not supported
        """
        pass

    def SetArchitectures(self, list_of_requested_architectures):
        """Confirms the requested architecture list is valid.

        TIP: Optional Override in a subclass

        Args:
            list_of_requested_architectures (list[str]): architectures to be built

        Raises:
            Exception: A requested architecture is not supported
        """
        pass

    def SetTargets(self, list_of_requested_target):
        """Confirms the requested target list is valid.

        TIP: Optional Override in a subclass

        Args:
            list_of_requested_targets (list[str]): targets to use

        Raises:
            Exception: A requested target is not supported
        """
        pass


class Edk2MultiPkgAwareInvocable(Edk2Invocable):
    """Base class for Multi-Pkg aware invocable.

    Attributes:
        requested_architecture_list (list): requested architectures to build
        requested_package_list (list): requested packages to build
        requested_target_list (list): requested targets to use

    TIP: Checkout Edk2Invocable Attributes
    to find any additional attributes that might exist.

    WARNING: This invocable should only be subclassed if creating a new invocable
    """

    def __init__(self):
        """Initializes the Invocable."""
        self.requested_architecture_list = []
        self.requested_package_list = []
        self.requested_target_list = []
        super().__init__()

    def AddCommandLineOptions(self, parserObj):
        """Adds command line options to the argparser."""
        # This will parse the packages that we are going to update
        parserObj.add_argument('-p', '--pkg', '--pkg-dir', dest='packageList', type=str,
                               help='Optional - A package or folder you want to update (workspace relative).'
                               'Can list multiple by doing -p <pkg1>,<pkg2> or -p <pkg3> -p <pkg4>',
                               action="append", default=[])
        parserObj.add_argument('-a', '--arch', dest="requested_arch", type=str, default=None,
                               help="Optional - CSV of architecutres requested to update. Example: -a X64,AARCH64")
        parserObj.add_argument('-t', '--target', dest='requested_target', type=str, default=None,
                               help="Optional - CSV of targets requested to update.  Example: -t DEBUG,NOOPT")

    def RetrieveCommandLineOptions(self, args):
        """Retrieve command line options from the argparser ."""
        packageListSet = set()
        for item in args.packageList:  # Parse out the individual packages
            item_list = item.split(",")
            for individual_item in item_list:
                # in case cmd line caller used Windows folder slashes
                individual_item = individual_item.replace("\\", "/").rstrip("/")
                packageListSet.add(individual_item.strip())
        self.requested_package_list = list(packageListSet)

        if args.requested_arch is not None:
            self.requested_architecture_list = args.requested_arch.upper().split(",")
        else:
            self.requested_architecture_list = []

        if args.requested_target is not None:
            self.requested_target_list = args.requested_target.upper().split(",")
        else:
            self.requested_target_list = []

    def InputParametersConfiguredCallback(self):
        """Initializes the environment once input parameters are collected."""
        if (len(self.requested_package_list) == 0):
            self.requested_package_list = list(self.PlatformSettings.GetPackagesSupported())
        self.PlatformSettings.SetPackages(self.requested_package_list)

        if (len(self.requested_architecture_list) == 0):
            self.requested_architecture_list = list(self.PlatformSettings.GetArchitecturesSupported())
        self.PlatformSettings.SetArchitectures(self.requested_architecture_list)

        if (len(self.requested_target_list) == 0):
            self.requested_target_list = list(self.PlatformSettings.GetTargetsSupported())
        self.PlatformSettings.SetTargets(self.requested_target_list)



# @file
#
# Command-line tool for inspecting UEFI Secure Boot databases
# Requires "pip install edk2-pytool-extensions"
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

"""Tool for inspecting UEFI Secure Boot databases.

The input file is a concatenations of EFI_SIGNATURE_LISTS, as is read by calling GetVariable() on
    Secure Boot variables (i.e. PK, KEK, db, dbx)
The output can be standard human-readable or compact for easier diffing
It can dump the database as-is, or a sorted & deduplicated canonical form, or just duplicates found in it
"""

import argparse
from edk2toollib.uefi.authenticated_variables_structure_support import EfiSignatureDatabase


def main():
    """Parses command-line parameters using ArgumentParser.

    Parses command-line parameters using ArgumentParser delegating to helper
    functions to fulfill the requests.
    """
    filenameHelp = 'Filename containing a UEFI Signature Database, \
        a concatenation of EFI_SIGNATURE_LISTs as read from GetVariable([PK, KEK, db, dbx])'

    sig_db_examples = '''
examples:

sig_db dump dbx_before.bin

sig_db --compact dump dbx_after.bin

sig_db --compact get_dupes dbx_with_dupes.bin

sig_db --compact get_canonical mixed_up_dbx.bin
'''

    parser = argparse.ArgumentParser(description='UEFI Signature database inspection tool',
                                     formatter_class=argparse.RawDescriptionHelpFormatter,
                                     epilog=sig_db_examples)

    parser.add_argument('--compact', action='store_true',
                        help='Compact, 1 line per data element output for easier diff-ing')

    subparsers = parser.add_subparsers(required=False, dest='action')

    parser_dump = subparsers.add_parser('dump', help='Print a UEFI Signature Database as-is in human-readable form')
    parser_dump.add_argument('file', type=str, help=filenameHelp)

    parser_get_dupes = subparsers.add_parser('get_dupes', help='Find duplicate signature entries in a UEFI Signature \
        Database. The test for duplication ignores SignatureOwner, testing only the SignatureData field. \
        Print them in UEFI Signature Database format, ordering is NOT maintained, output is NOT itself deduplicated')
    parser_get_dupes.add_argument('file', type=str, help='Filename of a UEFI Signature Database \
        (concatenation of EFI_SIGNATURE_LISTs as read from GetVariable() )')

    parser_get_canonical = subparsers.add_parser('get_canonical', help='Reduce a UEFI Signature Database to a \
        canonical (de-duplicated, sorted) form and print it')
    parser_get_canonical.add_argument('file', type=str,
                                      help='The name of the UEFI Signature Database file to get_canonical')

    options = parser.parse_args()

    if options.action is None:
        parser.print_help()
        return

    try:
        with open(options.file, 'rb') as f:
            esd = EfiSignatureDatabase(f)
            if (options.action == 'get_dupes'):
                esd = esd.GetDuplicates()
            elif (options.action == 'get_canonical'):
                esd = esd.GetCanonical()

            esd.Print(compact=options.compact)

    except FileNotFoundError:
        print('ERROR:  File not found: "{0}"'.format(options.file))


if __name__ == '__main__':
    main()



# @file test_edk2_ci_setup.py
# This contains unit tests for the edk2_ci_setup
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import unittest
from edk2toolext.invocables.edk2_ci_setup import Edk2CiBuildSetup
import sys
import shutil
import os
import logging
from importlib import reload
from edk2toolext.environment import shell_environment
from edk2toolext.tests.uefi_tree import uefi_tree
from edk2toolext.environment import self_describing_environment
from edk2toolext.environment import version_aggregator


class TestEdk2CiSetup(unittest.TestCase):

    minimalTree = None

    def setUp(self):
        TestEdk2CiSetup.restart_logging()
        tree = uefi_tree()
        self.minimalTree = tree.get_workspace()
        print(self.minimalTree)
        pass

    def tearDown(self):
        shell_environment.GetEnvironment().restore_initial_checkpoint()
        TestEdk2CiSetup.restart_logging()
        buildFolder = os.path.join(self.minimalTree, "Build")
        shutil.rmtree(buildFolder, ignore_errors=True)
        self_describing_environment.DestroyEnvironment()
        version_aggregator.ResetVersionAggregator()
        pass

    @classmethod
    def restart_logging(cls):
        '''
        We restart logging as logging is closed at the end of edk2 invocables.
        We also initialize it at the start.
        Reloading is the easiest way to get fresh state
        '''
        logging.shutdown()
        reload(logging)

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def test_init(self):
        builder = Edk2CiBuildSetup()
        self.assertIsNotNone(builder)

    def test_ci_setup(self):
        builder = Edk2CiBuildSetup()
        settings_file = os.path.join(self.minimalTree, "settings.py")
        sys.argv = ["stuart_ci_setup", "-c", settings_file, "-v"]
        try:
            builder.Invoke()
        except SystemExit as e:
            self.assertEqual(e.code, 0, "We should have a non zero error code")
            pass

    def test_ci_setup_bad_omnicache_path(self):
        builder = Edk2CiBuildSetup()
        settings_file = os.path.join(self.minimalTree, "settings.py")
        sys.argv = ["stuart_ci_setup", "-c", settings_file, "-v", "--omnicache", "does_not_exist"]
        try:
            builder.Invoke()
        except SystemExit as e:
            self.assertEqual(e.code, 0, "We should have a non zero error code")
            pass



# @file test_edk2_update.py
# This contains unit tests for the edk2_update
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import unittest
from edk2toolext.invocables.edk2_update import Edk2Update
import tempfile
import sys
import os
import logging
from importlib import reload
from edk2toolext.environment import shell_environment
from edk2toolext.tests.uefi_tree import uefi_tree
from edk2toolext.environment import self_describing_environment
from edk2toolext.environment import version_aggregator


class TestEdk2Update(unittest.TestCase):

    temp_folders = []

    def tearDown(self):
        shell_environment.GetEnvironment().restore_initial_checkpoint()
        for temp_folder in TestEdk2Update.temp_folders:
            logging.info(f"Cleaning up {temp_folder}")
            # shutil.rmtree(os.path.abspath(temp_folder), ignore_errors=True)
        TestEdk2Update.restart_logging()
        # we need to make sure to tear down the version aggregator and the SDE
        self_describing_environment.DestroyEnvironment()
        version_aggregator.ResetVersionAggregator()

    @classmethod
    def restart_logging(cls):
        logging.shutdown()
        reload(logging)

    @classmethod
    def updateClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    @classmethod
    def get_temp_folder(cls):
        temp_folder = os.path.abspath(tempfile.mkdtemp())
        TestEdk2Update.temp_folders.append(temp_folder)
        return os.path.abspath(temp_folder)

    def invoke_update(self, settings_filepath, args=[], failure_expected=False):
        sys.argv = ["stuart_update", "-c", settings_filepath]
        sys.argv.extend(args)
        builder = Edk2Update()
        try:
            builder.Invoke()
        except SystemExit as e:
            if failure_expected:
                self.assertIsNot(e.code, 0, "We should have a non zero error code")
            else:
                self.assertIs(e.code, 0, "We should have a zero error code")
        return builder

    #######################################
    # Test methods
    def test_init(self):
        WORKSPACE = self.get_temp_folder()
        tree = uefi_tree(WORKSPACE)
        settings_filepath = tree.get_settings_provider_path()
        sys.argv = ["stuart_update", "-c", settings_filepath]
        builder = Edk2Update()
        self.assertIsNotNone(builder)

    def test_one_level_recursive(self):
        ''' makes sure we can do a recursive update '''
        WORKSPACE = self.get_temp_folder()
        tree = uefi_tree(WORKSPACE)
        logging.getLogger().setLevel(logging.WARNING)
        tree.create_Edk2TestUpdate_ext_dep()
        # Do the update
        updater = self.invoke_update(tree.get_settings_provider_path())
        # make sure it worked
        self.assertTrue(os.path.exists(os.path.join(WORKSPACE, "Edk2TestUpdate_extdep",
                                                    "NuGet.CommandLine_extdep", "extdep_state.json")))
        build_env, shell_env, failure = updater.PerformUpdate()
        # we should have no failures
        self.assertEqual(failure, 0)
        # we should have found two ext deps
        self.assertEqual(len(build_env.extdeps), 2)

    def test_multiple_extdeps(self):
        ''' makes sure we can do multiple ext_deps at the same time '''
        WORKSPACE = self.get_temp_folder()
        tree = uefi_tree(WORKSPACE)
        num_of_ext_deps = 5
        logging.getLogger().setLevel(logging.WARNING)
        tree.create_ext_dep("nuget", "NuGet.CommandLine", "5.2.0")
        tree.create_ext_dep("nuget", "NuGet.LibraryModel", "5.6.0")
        tree.create_ext_dep("nuget", "NuGet.Versioning", "5.6.0")
        tree.create_ext_dep("nuget", "NuGet.Packaging.Core", "5.6.0")
        tree.create_ext_dep("nuget", "NuGet.RuntimeModel", "4.2.0")
        # Do the update
        updater = self.invoke_update(tree.get_settings_provider_path())
        build_env, shell_env, failure = updater.PerformUpdate()
        # we should have no failures
        self.assertEqual(failure, 0)
        # we should have found two ext deps
        self.assertEqual(len(build_env.extdeps), num_of_ext_deps)

    def test_duplicate_ext_deps(self):
        ''' verifies redundant ext_deps fail '''
        WORKSPACE = self.get_temp_folder()
        tree = uefi_tree(WORKSPACE)

        logging.getLogger().setLevel(logging.WARNING)

        tree.create_ext_dep(dep_type="nuget",
                            name="NuGet.CommandLine",
                            version="5.2.0",
                            dir_path="1",
                            extra_data={"id:": "CmdLine1"})
        tree.create_ext_dep(dep_type="nuget",
                            name="NuGet.CommandLine",
                            version="5.2.0",
                            dir_path="2",
                            extra_data={"id:": "CmdLine1"})

        # Do the update. Expect a ValueError from the version aggregator.
        with self.assertRaises(ValueError):
            self.invoke_update(tree.get_settings_provider_path(), failure_expected=True)

    def test_duplicate_ext_deps_skip_dir(self):
        ''' verifies redundant ext_deps pass if one is skipped '''
        WORKSPACE = self.get_temp_folder()
        tree = uefi_tree(WORKSPACE)
        num_of_ext_deps = 1

        logging.getLogger().setLevel(logging.WARNING)

        tree.create_ext_dep(dep_type="nuget",
                            name="NuGet.CommandLine",
                            version="5.2.0",
                            dir_path="1",
                            extra_data={"id:": "CmdLine1"})
        tree.create_ext_dep(dep_type="nuget",
                            name="NuGet.CommandLine",
                            version="5.2.0",
                            dir_path="2",
                            extra_data={"id:": "CmdLine1"})

        # Update GetSkippedDirectories() implementation
        with open(tree.get_settings_provider_path(), 'r') as s:
            settings_text = s.read()

        settings_text = settings_text.replace(
            'def GetSkippedDirectories(self):\n        return ()',
            'def GetSkippedDirectories(self):\n        return (\"2\",)')

        with open(tree.get_settings_provider_path(), 'w') as s:
            s.write(settings_text)

        # Do the update
        updater = self.invoke_update(tree.get_settings_provider_path())
        build_env, shell_env, failure = updater.PerformUpdate()
        # we should have found one ext dep
        self.assertEqual(len(build_env.extdeps), num_of_ext_deps)
        # the one ext_dep should be valid
        self.assertEqual(failure, 0)

    def test_multiple_duplicate_ext_deps_skip_dir(self):
        ''' verifies multiple ext_deps in sub dirs are skipped'''
        WORKSPACE = self.get_temp_folder()
        tree = uefi_tree(WORKSPACE)
        num_of_ext_deps = 1

        logging.getLogger().setLevel(logging.WARNING)

        tree.create_ext_dep(dep_type="nuget",
                            name="NuGet.CommandLine",
                            version="5.2.0",
                            dir_path="first/second",
                            extra_data={"id:": "CmdLine1"})
        tree.create_ext_dep(dep_type="nuget",
                            name="NuGet.CommandLine",
                            version="5.2.0",
                            dir_path="third/fourth/fifth",
                            extra_data={"id:": "CmdLine1"})
        tree.create_ext_dep(dep_type="nuget",
                            name="NuGet.CommandLine",
                            version="5.2.0",
                            dir_path="sixth/seventh/eighth",
                            extra_data={"id:": "CmdLine1"})

        # Update GetSkippedDirectories() implementation
        with open(tree.get_settings_provider_path(), 'r') as s:
            settings_text = s.read()

        settings_text = settings_text.replace(
            'def GetSkippedDirectories(self):\n        return ()',
            'def GetSkippedDirectories(self):\n        return (\"third\",\"sixth\")')

        with open(tree.get_settings_provider_path(), 'w') as s:
            s.write(settings_text)

        # Do the update
        updater = self.invoke_update(tree.get_settings_provider_path())
        build_env, shell_env, failure = updater.PerformUpdate()
        # we should have found one ext dep
        self.assertEqual(len(build_env.extdeps), num_of_ext_deps)
        # the one ext_dep should be valid
        self.assertEqual(failure, 0)

    def test_bad_ext_dep(self):
        ''' makes sure we can do an update that will fail '''
        WORKSPACE = self.get_temp_folder()
        tree = uefi_tree(WORKSPACE)
        logging.getLogger().setLevel(logging.WARNING)
        # we know this version is bad
        tree.create_Edk2TestUpdate_ext_dep("0.0.0")
        # Do the update
        updater = self.invoke_update(tree.get_settings_provider_path(), failure_expected=True)
        build_env, shell_env, failure = updater.PerformUpdate()
        # we should have no failures
        self.assertEqual(failure, 1)



## @file signing_helper_test.py
# This unittest module contains test cases for the signing_helper module.
#
##
# Copyright (C) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##


import os
import unittest

from edk2toolext.capsule import signing_helper


class SignerLocationTests(unittest.TestCase):
    def test_should_be_able_to_fetch_a_builtin_signer_module(self):
        py_signer = signing_helper.get_signer(signing_helper.PYOPENSSL_SIGNER)
        self.assertTrue(hasattr(py_signer, 'sign'))

        signtoolsigner = signing_helper.get_signer(signing_helper.SIGNTOOL_SIGNER)
        self.assertTrue(hasattr(signtoolsigner, 'sign'))

    def test_should_be_able_to_pass_a_signing_module(self):
        py_signer = signing_helper.get_signer(
            signing_helper.PYPATH_MODULE_SIGNER,
            'edk2toolext.capsule.pyopenssl_signer'
        )
        self.assertTrue(hasattr(py_signer, 'sign'))

    def test_should_be_able_to_fetch_a_user_provided_signer_module(self):
        test_script_path = os.path.dirname(os.path.abspath(__file__))
        # TEST NOTE: This is a little hacky and will break if relative paths are changed.
        module_root_path = os.path.dirname(os.path.dirname(test_script_path))
        py_signer_path = os.path.join(module_root_path, 'capsule', 'pyopenssl_signer.py')
        self.assertTrue(os.path.isfile(py_signer_path))
        py_signer = signing_helper.get_signer(signing_helper.LOCAL_MODULE_SIGNER, py_signer_path)
        self.assertTrue(hasattr(py_signer, 'sign'))

# NOTE: These tests may not run on non-Windows or without the WDK installed.
# class SigntoolSignerModuleTest(unittest.TestCase):
#     def test_module_should_be_able_to_locate_signtool(self):
#         signtoolsigner = signing_helper.get_signer(signing_helper.SIGNTOOL_SIGNER)
#         self.assertTrue(os.path.isfile(signtoolsigner.get_signtool_path()))



# @file test_git_dependency.py
# Unit test suite for the GitDependency class.
#
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import os
import unittest
import logging
import tempfile
import copy
from edk2toollib.utility_functions import RemoveTree
from edk2toolext.environment.external_dependency import ExternalDependency as ExtDep

TEST_DIR = None
GOOD_VERSION = "5.2.0"
BAD_VERSION = "5.2.13.1.2"
MISSING_VERSION = "5.200.13"

NUGET_TEMPLATE = {
    "scope": "global",
    "type": "nuget",
    "name": "NuGet.CommandLine",
    "source": "https://api.nuget.org/v3/index.json",
}

WEB_TEMPLATE = {
    "scope": "global",
    "type": "web",
    "name": "NuGet.CommandLine",
    "source": "https://example.com/some_other_site",
}


def prep_workspace():
    global TEST_DIR
    # if test temp dir doesn't exist
    if TEST_DIR is None or not os.path.isdir(TEST_DIR):
        TEST_DIR = tempfile.mkdtemp()
        logging.debug("temp dir is: %s" % TEST_DIR)
    else:
        clean_workspace()
        TEST_DIR = tempfile.mkdtemp()


def clean_workspace():
    global TEST_DIR
    if TEST_DIR is None:
        return

    if os.path.isdir(TEST_DIR):
        RemoveTree(TEST_DIR)
        TEST_DIR = None


class TestExternalDependency(unittest.TestCase):
    def setUp(self):
        prep_workspace()

    @classmethod
    def setUpClass(cls):
        logger = logging.getLogger('')
        logger.addHandler(logging.NullHandler())
        unittest.installHandler()

    @classmethod
    def tearDownClass(cls):
        clean_workspace()

    def test_determine_cache_path(self):
        nuget_desc = copy.copy(NUGET_TEMPLATE)
        nuget_desc['version'] = GOOD_VERSION
        nuget_desc['descriptor_file'] = os.path.join(TEST_DIR, 'non_file.yaml')
        ext_dep = ExtDep(nuget_desc)

        cache_path = os.path.join(TEST_DIR, 'stuart_cache')

        self.assertIsNone(ext_dep.determine_cache_path())

        ext_dep.set_global_cache_path(cache_path)
        self.assertIsNone(ext_dep.determine_cache_path())

        os.makedirs(cache_path)
        self.assertTrue(ext_dep.determine_cache_path().startswith(os.path.join(cache_path, 'nuget')))

        web_desc = copy.copy(WEB_TEMPLATE)
        web_desc['version'] = GOOD_VERSION
        web_desc['descriptor_file'] = os.path.join(TEST_DIR, 'non_file.yaml')
        ext_dep = ExtDep(web_desc)
        ext_dep.set_global_cache_path(cache_path)
        self.assertTrue(ext_dep.determine_cache_path().startswith(os.path.join(cache_path, 'web')))

    def test_different_versions_have_different_caches(self):
        cache_path = os.path.join(TEST_DIR, 'stuart_cache')
        os.makedirs(cache_path)

        nuget_desc1 = copy.copy(NUGET_TEMPLATE)
        nuget_desc1['version'] = GOOD_VERSION
        nuget_desc1['descriptor_file'] = os.path.join(TEST_DIR, 'non_file.yaml')
        ext_dep1 = ExtDep(nuget_desc1)

        nuget_desc2 = copy.copy(NUGET_TEMPLATE)
        nuget_desc2['version'] = "7.0.0"
        nuget_desc2['descriptor_file'] = os.path.join(TEST_DIR, 'non_file.yaml')
        ext_dep2 = ExtDep(nuget_desc2)

        ext_dep1.set_global_cache_path(cache_path)
        ext_dep2.set_global_cache_path(cache_path)

        self.assertNotEqual(ext_dep1.determine_cache_path(), ext_dep2.determine_cache_path())

    def test_different_sources_have_different_caches(self):
        cache_path = os.path.join(TEST_DIR, 'stuart_cache')
        os.makedirs(cache_path)

        nuget_desc1 = copy.copy(NUGET_TEMPLATE)
        nuget_desc1['version'] = GOOD_VERSION
        nuget_desc1['descriptor_file'] = os.path.join(TEST_DIR, 'non_file.yaml')
        ext_dep1 = ExtDep(nuget_desc1)

        nuget_desc2 = copy.copy(NUGET_TEMPLATE)
        nuget_desc2['version'] = GOOD_VERSION
        nuget_desc2['descriptor_file'] = os.path.join(TEST_DIR, 'non_file.yaml')
        nuget_desc2['source'] = "https://api.nuget.org/v3/different_index.json"
        ext_dep2 = ExtDep(nuget_desc2)

        ext_dep1.set_global_cache_path(cache_path)
        ext_dep2.set_global_cache_path(cache_path)

        self.assertNotEqual(ext_dep1.determine_cache_path(), ext_dep2.determine_cache_path())

    def test_can_copy_to_cache(self):
        cache_path = os.path.join(TEST_DIR, 'stuart_cache')
        os.makedirs(cache_path)

        nuget_desc = copy.copy(NUGET_TEMPLATE)
        nuget_desc['version'] = GOOD_VERSION
        nuget_desc['descriptor_file'] = os.path.join(TEST_DIR, 'non_file.yaml')
        ext_dep = ExtDep(nuget_desc)
        ext_dep.set_global_cache_path(cache_path)

        self.assertFalse(os.path.exists(ext_dep.determine_cache_path()))

        # Create a new directory with a dummy file.
        test_path = os.path.join(TEST_DIR, 'test_path')
        test_file = os.path.join(test_path, 'test_file.txt')
        os.makedirs(test_path)
        with open(test_file, "w") as fp:
            fp.write("DEADBEEF\n")

        # Copy the test data to the cache.
        ext_dep.copy_to_global_cache(test_path)

        self.assertTrue(os.path.exists(ext_dep.determine_cache_path()))
        copied_file = os.path.join(ext_dep.determine_cache_path(), 'test_file.txt')
        self.assertTrue(os.path.exists(copied_file))
        file_contents = None
        with open(copied_file, "r") as fp:
            file_contents = fp.read()
        self.assertTrue("DEADBEEF" in file_contents)

    def test_can_copy_from_cache(self):
        cache_path = os.path.join(TEST_DIR, 'stuart_cache')
        os.makedirs(cache_path)

        nuget_desc = copy.copy(NUGET_TEMPLATE)
        nuget_desc['version'] = GOOD_VERSION
        nuget_desc['descriptor_file'] = os.path.join(TEST_DIR, 'non_file.yaml')
        ext_dep = ExtDep(nuget_desc)
        ext_dep.set_global_cache_path(cache_path)

        self.assertFalse(os.path.exists(ext_dep.determine_cache_path()))

        test_path = os.path.join(TEST_DIR, 'test_path')
        test_file = os.path.join(test_path, 'test_file.txt')
        os.makedirs(test_path)
        with open(test_file, "w") as fp:
            fp.write("DEADBEEF\n")
        ext_dep.copy_to_global_cache(test_path)

        test_path2 = os.path.join(TEST_DIR, 'test_path2')
        self.assertFalse(os.path.exists(test_path2))

        ext_dep.copy_from_global_cache(test_path2)

        self.assertTrue(os.path.exists(test_path2))
        copied_file = os.path.join(test_path2, 'test_file.txt')
        self.assertTrue(os.path.exists(copied_file))
        file_contents = None
        with open(copied_file, "r") as fp:
            file_contents = fp.read()
        self.assertTrue("DEADBEEF" in file_contents)


if __name__ == '__main__':
    unittest.main()



# @file pyopenssl_signer.py
# This module contains the abstracted signing interface for pyopenssl. This interface
# abstraction takes in the signature_options and signer_options dictionaries that are
# used by capsule_tool and capsule_helper.
#
##
# Copyright (C) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Abstracted signing interface for pyopenssl.

This interface abstraction takes in the signature_options and signer_options
dictionaries that are used by capsule_tool and capsule_helper.
"""
import logging
import warnings

from OpenSSL import crypto


def sign(data: bytes, signature_options: dict, signer_options: dict) -> bytes:
    """Primary signing interface.

    Takes in the signature_options and signer_options
    dictionaries that are used by capsule_tool and capsule_helper.

    Args:
        data (bytes): data to write into the sign tool.
        signature_options (dict): dictionary containing signature options
        signer_options (dict): dictionary containing signer options

    Raises:
        (ValueError()): Unsupported signature or signer options
        (RuntimeError()): Signtool.exe returned with error

    NOTE: Currently, we only support the necessary algorithms for capsules.
    """
    # The following _if_ clause handles the deprecated signature_option 'sign_alg' for backwards compatibility
    # when the deprecated option is supplied, this code adds the new, required options based on prior code behavior
    if 'sign_alg' in signature_options:
        warnings.warn('Signature_option "sign_alg" is deprecated, use "type"', DeprecationWarning)
        if signature_options['sign_alg'] == 'pkcs12':
            # map legacy behavior to new options and backwards-compatible values
            signature_options['type'] = 'bare'
            signature_options['encoding'] = 'binary'
            signer_options['key_file_format'] = 'pkcs12'
        else:
            raise ValueError(f"Unsupported signature algorithm: {signature_options['sign_alg']}!")

    ''' signature type 'bare' is just a binary signed digest, no PEM headers/footers or ASN '''
    if signature_options['type'] != 'bare':
        raise ValueError(f"Unsupported signature type: {signature_options['type']}")
    if 'type_options' in signature_options:
        raise ValueError("Signature type options not supported")
    if signature_options['encoding'] != 'binary':
        raise ValueError(f"Unsupported signature encoding: {signature_options['encoding']}")
    if signature_options['hash_alg'] != 'sha256':
        raise ValueError(f"Unsupported hashing algorithm: {signature_options['hash_alg']}")
    if signer_options['key_file_format'] != 'pkcs12':
        raise ValueError(f"Unsupported signer key file format: {signer_options['key_file_format']}")

    logging.debug("Executing PKCS1 Signing")

    # If a key file is provided, use it for signing.
    if 'key_file' in signer_options:
        with open(signer_options['key_file'], 'rb') as key_file:
            signer_options['key_data'] = key_file.read()

    # TODO: Figure out OIDs.
    # TODO: Figure out EKU.

    pkcs12 = crypto.load_pkcs12(signer_options['key_data'])
    return crypto.sign(pkcs12.get_privatekey(), data, signature_options['hash_alg'])



## @file capsule_tool_test.py
# This unittest module contains test cases for the capsule_tool module and CLI routines.
#
##
# Copyright (C) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##


import os
import unittest
import tempfile
import json
import yaml

from edk2toolext.capsule import capsule_tool

DUMMY_OPTIONS = {
    'capsule': {
        'option1': 'value1'
    },
    'signer': {
        'option2': 'value2',
        'option_not': 'orig_value'
    }
}
DUMMY_OPTIONS_FILE_NAME = 'dummy_options_file'
DUMMY_PAYLOAD_FILE_NAME = 'dummy_payload'


class ParameterParsingTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # We'll use the one-time setup to create
        # any temporary test files we'll need.
        cls.temp_dir = tempfile.mkdtemp()
        cls.dummy_json_options = os.path.join(cls.temp_dir, DUMMY_OPTIONS_FILE_NAME + ".json")
        cls.dummy_yaml_options = os.path.join(cls.temp_dir, DUMMY_OPTIONS_FILE_NAME + ".yaml")
        cls.dummy_payload = os.path.join(cls.temp_dir, DUMMY_PAYLOAD_FILE_NAME + ".bin")

        with open(cls.dummy_json_options, 'w') as dummy_file:
            json.dump(DUMMY_OPTIONS, dummy_file)
        with open(cls.dummy_yaml_options, 'w') as dummy_file:
            yaml.dump(DUMMY_OPTIONS, dummy_file)
        with open(cls.dummy_payload, 'wb') as dummy_file:
            dummy_file.write(b'DEADBEEF')

    @unittest.skip("test is incomplete")
    def test_should_require_a_signer_option(self):
        pass

    def test_capsule_options_should_be_passable(self):
        cli_params = ['--builtin_signer', 'pyopenssl']
        parsed_args = capsule_tool.get_cli_options(cli_params + [self.dummy_payload, self.temp_dir])
        self.assertEqual(len(parsed_args.capsule_options), 0)

        cli_params += ['-dc', 'option1=value1']
        cli_params += ['-dc', 'option2=value2']
        cli_params += [self.dummy_payload, self.temp_dir]
        parsed_args = capsule_tool.get_cli_options(cli_params)
        self.assertEqual(len(parsed_args.capsule_options), 2)

    def test_signer_options_should_be_passable(self):
        cli_params = ['--builtin_signer', 'pyopenssl']
        parsed_args = capsule_tool.get_cli_options(cli_params + [self.dummy_payload, self.temp_dir])
        self.assertEqual(len(parsed_args.signer_options), 0)

        cli_params += ['-ds', 'option1=value1']
        cli_params += ['-ds', 'option2=value2']
        cli_params += [self.dummy_payload, self.temp_dir]
        parsed_args = capsule_tool.get_cli_options(cli_params)
        self.assertEqual(len(parsed_args.signer_options), 2)

    def test_should_not_accept_an_invalid_path(self):
        cli_params = ['--builtin_signer', 'pyopenssl']
        cli_params += ['-o', 'not_a_path.bin']
        cli_params += [self.dummy_payload]
        with self.assertRaises(SystemExit):
            capsule_tool.get_cli_options(cli_params)

    def test_should_not_load_an_invalid_path(self):
        bad_path = 'not_a_path.bin'
        loaded_options = capsule_tool.load_options_file(bad_path)
        self.assertEqual(loaded_options, None)

    def test_options_file_should_load_json(self):
        with open(self.dummy_json_options, 'r') as options_file:
            loaded_options = capsule_tool.load_options_file(options_file)

        self.assertEqual(loaded_options['capsule']['option1'], 'value1')
        self.assertEqual(loaded_options['signer']['option2'], 'value2')

    def test_options_file_should_load_yaml(self):
        with open(self.dummy_yaml_options, 'r') as options_file:
            loaded_options = capsule_tool.load_options_file(options_file)

        self.assertEqual(loaded_options['capsule']['option1'], 'value1')
        self.assertEqual(loaded_options['signer']['option2'], 'value2')

    # @pytest.mark.skip(reason="test is incomplete")
    def test_cli_options_should_override_file_options(self):
        capsule_cli_options = ['option1=value2', 'new_option=value3']
        signer_cli_options = ['option2=value7', 'option2=value8']

        final_options = capsule_tool.update_options(DUMMY_OPTIONS, capsule_cli_options, signer_cli_options)

        self.assertEqual(final_options['capsule']['option1'], 'value2')
        self.assertEqual(final_options['capsule']['new_option'], 'value3')
        self.assertEqual(final_options['signer']['option2'], 'value8')
        self.assertEqual(final_options['signer']['option_not'], 'orig_value')

    def test_full_options_path_should_work(self):
        # Parse the command parameters.
        cli_params = ['--builtin_signer', 'pyopenssl']
        cli_params += ['-o', self.dummy_json_options]
        cli_params += ['-dc', 'option1=value2']
        cli_params += ['-dc', 'new_option=value3']
        cli_params += ['-ds', 'option2=value7']
        cli_params += ['-ds', 'option2=value8']
        cli_params += [self.dummy_payload, self.temp_dir]
        parsed_args = capsule_tool.get_cli_options(cli_params)

        loaded_options = capsule_tool.load_options_file(parsed_args.options_file)
        final_options = capsule_tool.update_options(
            loaded_options,
            parsed_args.capsule_options,
            parsed_args.signer_options
        )

        self.assertEqual(final_options['capsule']['option1'], 'value2')
        self.assertEqual(final_options['capsule']['new_option'], 'value3')
        self.assertEqual(final_options['signer']['option2'], 'value8')
        self.assertEqual(final_options['signer']['option_not'], 'orig_value')
